# -*- coding: utf-8 -*-
"""
Created on Mon May 15 15:41:41 2023

@author: romain.coulon
"""



import zipfile
import numpy as np

file = 'decayData/All-nuclides_Ensdf.zip'
z = zipfile.ZipFile(file)

with z.open('Ag-108m.txt') as f:
    data = f.readlines()
    nL = np.size(data)
    
    ## fomatage en list de vecteur pour chaque line du fichier ENSDF
    for i in range(nL):              # boucle dans data
        data[i] = str(data[i])                  # conversion en string
        data[i] = data[i].replace("b'",'')       # on enlève les caratères de formatage
        data[i] = data[i].replace("\\r\\n'",'')  # on enlève les caratères de formatage
    for i in range(nL):
        data[i] = data[i].split()
    
    for i in range(nL):
        if i>0 and ("L" in data[i]) and ("AUGER" in data[i]) and ("|]" in data[i-1]):
            data.insert(i,[data[i][0], 'T'])

    ## Définir des repères
    index_auger = []          # indice de
    daugther = []             # liste des filles
    index_end = []            # indice dernier valeur pour une fille
    posi = []                 # indice des blocs transitions 
    for i, p in enumerate(data): # boucle dans le fichier
        if 'DECAY' in p:
            daugther.append(p[0])   # liste des filles
        if 'Auger' in p:
            index_auger.append(i)   # repère de ligne pour ensuite définir l'ensemble des blocs
        if len(p)==2:
            posi.append(i)          # repère de ligne pour ensuite définir les blocs (2)
        # if 'L' in p
        #     posi:
        if 'P' in p:
            index_end.append(i)     # repère de fin pour ensuite définir l'ensemble des blocs
            posi.append(i)
            
    ## Filtrage des données utiles et formatage
    energy = []
    prob = []
    Type = []
    Fille = []
    for i in range(len(posi)-1): # bloucle dans les blocs
        start = posi[i]+1        # indice debut du bloc
        end = posi[i+1]          # indice fin du bloc
        d = data[start:end]      # Le bloc
        e_b = []                   # Energie des transitions du bloc
        prob_b = []                # Proba des transitions du bloc
        Type_b = []                # Types des transitions du bloc
        Fille_b = []               # Fille des transitions du bloc
        #e2 = []
        if start == end:
            continue
        if start-1 in index_end:
            #e = []
            continue
        for n, p1 in enumerate(d):
            #e2 = []
            if '-' in p1[2]: # traitement des intervals de nombres réels
                x = p1[2].split('-')
                p1[2] = round((float(x[1])+float(x[0]))/2,3) # Moyenne
            if '|]' in p1:        # traitment des accolades 
                if len(p1)>6:     # repère de la ligne portant la proba pour le groupe 
                    prob_b.append(float(p1[4]))
                e_b.append(float(p1[2]))
            else:
                e_b.append(float(p1[2]))
                prob_b.append(float(p1[3]))
        
        if len(prob_b)==1 and len(e_b)>1:
            e_b = [np.mean(e_b)]
        
        print(e_b)
        print(prob_b)
        print(" ")



# -*- coding: utf-8 -*-
"""
Created on Wed Jul  5 10:04:53 2023

@author: romain.coulon
"""

import numpy as np
import TDCRPy as td
import scipy.optimize as opt
import sys, time
sys.path.insert(1, 'G:\Python_modules\BIPM_RI_PyModules')
import TDCRcalculation as tc 



def eff(TDCR_measure, TAB, TBC, TAC, Rad, kB):
    
    N=50000
    L=1
    # r=opt.minimize_scalar(td.TDCRPy, args=(TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "res", "sym"), method='golden')#,options={'xatol': 1e-7, 'disp': True, 'maxiter':100})
    # r=opt.minimize_scalar(td.TDCRPy, args=(TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "res", "sym"), method='brent')
    r=opt.minimize_scalar(td.TDCRPy, args=(TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "res", "sym"), method='bounded', bounds=[0.5, 2])
    
    # r=opt.minimize(td.TDCRPy, L, args=(TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "res", "sym"), method='nelder-mead',options={'xatol': 1e-7, 'disp': True, 'maxiter':100})
    L=r.x
    print(r)
    #L=(L*0.995, L*1.021, L*0.988)           # Free paramete in keV-1
    # L = (L, L, L)
    # r=opt.minimize(td.TDCRPy, L, args=(TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "res", "asym"), method='nelder-mead',options={'xatol': 1e-7, 'disp': True, 'maxiter':100})
    # L=r.x
    # print(r)
    out=td.TDCRPy(L,TDCR_measure, TAB, TBC, TAC, Rad, "1", N, kB, 0.98, 1000, "eff", "sym")
    return np.mean(L), L, out[2], out[2], out[3]
    
def TicTocGenerator():
    """
    Generator that returns time differences
    """
    ti = 0           # initial time
    tf = time.time() # final time
    while True:
        ti = tf
        tf = time.time()
        yield tf-ti # returns the time difference

TicToc = TicTocGenerator() # create an instance of the TicTocGen generator

# This will be the main function through which we define both tic() and toc()
def toc(tempBool=True):
    """
    Prints the time difference yielded by generator instance TicToc
    """
    
    tempTimeInterval = next(TicToc)
    if tempBool:
        print( "Elapsed time: %f seconds.\n" %tempTimeInterval )

def tic():
    """
    Records a time in TicToc, marks the beginning of a time interval
    """
    toc(False)


"""
TEST
"""

Rad="Co-60"    # list of radionuclides (Na-24)
kB =1.0e-5       # Birks constant in cm/keV  
TDCR_measure = 0.977667386529166        # Measured TDCR value
TAB = 0.992232838598821
TBC = 0.992343419459002
TAC = 0.99275350064608



tic()
F1, FFF, eff1, eff2 = tc.I2calc(TDCR_measure, TAB, TBC, TAC, Rad, kB)
toc()

print("/nanalytical model")
print("free parameter = ", F1)
print("free parameters = ", FFF)
print("Double count rate efficiency (sym) = ", eff1)
print("Double count rate efficiency (asym) = ", eff2)


tic()
F1, FFF, eff1, eff2, u = eff(TDCR_measure, TAB, TBC, TAC, Rad, kB)
toc()

print("/nstochastic model")
print("free parameter = ", F1)
print("free parameters = ", FFF)
print("Double count rate efficiency (sym) = ", eff1)
print("Double count rate efficiency (asym) = ", eff2)
print("u(eff) = ", u)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 16:01:49 2023

A Monte-Carlo code to calculate detection efficiency in TDCR measurements

@author: Romain Coulon, Jialin Hu
Bureau International des Poids et Mesures
"""

## IMPORT PYTHON MODULES
import tdcrpy.TDCR_model_lib as tl
import numpy as np

def TDCRPy(L, TD, TAB, TBC, TAC, Rad, pmf_1, N, kB, RHO, nE, mode, mode2, Display=False):
    """
    This is a Monte-Carlo TDCR model

    Parameters
    ----------
    L : Float (if mode2="sym") or a tuple (if mode2="asym")
        Free parameter.
    TD : Float
        Measured TDCR.
    Rad : string
        List of radionuclides.
    pmf_1 : string
        list of probability of each radionuclide.
    N : integer
        Number of Monte-Carlo trials.
    kB : float
        Birks constant.
    RHO : float
        Density of the scintillator.
    nE : integer
        Number of bins for the quenching function.
    mode : string
        "res" to return the residual, "eff" to return efficiencies.
    mode2 : string
        "sym" for symetrical model, "asym" for symetrical model.
    Display : Boolean, optional
        "True" to display details on the decay sampling. The default is False.

    Returns
    -------
    List
        if mode=="res", the residual (float).
        if mode=="eff", the efficiencies (list)

    """
    
    
    Rad=Rad.replace(" ","")
    Rad=Rad.split(",")
    pmf_1=pmf_1.split(",")
    pmf_1 = [float(x) for x in pmf_1]
    
    if np.size(pmf_1) > 1:
        if sum(pmf_1) !=1: print("warning p not equal to 1")
    elif pmf_1[0] != 1: print("warning")

    """
    Read PenNuc File
    """
    out_PenNuc = []   
    particle = []          # Particle(s) from the Mother  --  indice 3
    p_branch = []          # Probablity of the different decay of branch -- indice 5
    e_branch = []          # Energy of the different decay of branch -- indice 4
    LevelDaughter = []     # Level of the Daughter nucleus just after the particle emission -- indice 6
    levelNumber = []       # The vector of level of the daughter to get information of all possible isomeric transitions -- indice 11
    prob_trans = []        # Probability for each transition -- indice 10
    prob_branch = []       # Probability for each branch -- indice 7
    levelEnergy = []       # Energy of each level -- indice 13
    transitionType = []    # type of each possible transitions (internal transitions or gamma emission) -- indice 8
    e_trans = []           # Energy of the transition -- indice 9
    next_level = []        # Next level on the daughter nucleus -- indice 12
    Q_value = []           # Energy of the reaction -- indice 2
    DaughterVec = []       # Daughters -- indice 0
    Pdaughter = []         # Probabiblity related to daughters -- indice 1
    Transition_prob_sum = []

    for rad_i in Rad:
        out_PenNuc = tl.readPenNuc2(rad_i)
        particle.append(out_PenNuc[3])       # Particle(s) from the Mother  --  indice 3
        p_branch.append(out_PenNuc[5])       # Probablity of the different decay of branch -- indice 5
        e_branch.append(out_PenNuc[4])       # Energy of the different decay of branch -- indice 4
        LevelDaughter.append(out_PenNuc[6])  # Level of the Daughter nucleus just after the particle emission -- indice 6
        levelNumber.append(out_PenNuc[11])   # The vector of level of the daughter to get information of all possible isomeric transitions -- indice 11
        prob_trans.append(out_PenNuc[10])    # Probability for each transition -- indice 10
        prob_branch.append(out_PenNuc[7])    # Probability for each branch -- indice 7
        levelEnergy.append(out_PenNuc[13])   # Energy of each level -- indice 13
        transitionType.append(out_PenNuc[8]) # type of each possible transitions (internal transitions or gamma emission) -- indice 8
        e_trans.append(out_PenNuc[9])        # Energy of the transition -- indice 9
        next_level.append(out_PenNuc[12])    # Next level on the daughter nucleus -- indice 12
        Q_value.append(out_PenNuc[2])        # Energy of the reaction -- indice 2
        DaughterVec.append(out_PenNuc[0])    # Daughters -- indice 0
        Pdaughter.append(out_PenNuc[1])      # Probabiblity related to daughters -- indice 1
        Transition_prob_sum.append(out_PenNuc[14])

        efficiency_S = []
        efficiency_D = []
        efficiency_T = []
        efficiency_AB = []
        efficiency_BC = []
        efficiency_AC = []
    for i in range(N): # Main Loop - Monte Carlo trials
        particle_vec=[]
        energy_vec=[]
        '''
        ===============================
        0. SAMPLING OF THE RADIONUCLIDE
        ===============================
        '''
        index_rad = tl.sampling(pmf_1)
        rad_i = Rad[index_rad]
        if Display: print("\n Sampled radionuclide = ", rad_i, "- L = ", L, ' keV-1 - kB = ', kB, ' cm/keV')

        '''
        ===========================
        I. DESINTEGRATION NUCLEAIRE
        ===========================
        '''
        #=========================
        # Sampling of the daughter
        #=========================
        iDaughter=tl.sampling(np.asarray(Pdaughter[index_rad])/sum(np.asarray(Pdaughter[index_rad])))
        Daughter = DaughterVec[index_rad][iDaughter]
        if Display: print("\t Sampled daughter:")
        if Display: print("\t\t Daughter = ", Daughter)           
        #=============================
        # Sampling of the decay branch
        #=============================
        branch_i = tl.normalise(prob_branch[index_rad][iDaughter])   # normalise la proba de branch
        i_branch=tl.sampling(branch_i)                               # indice de la branche globale
        if p_branch[index_rad][iDaughter][i_branch] != []:
            branch_proba = tl.normalise(p_branch[index_rad][iDaughter][i_branch])
            index_subBranch = tl.sampling(branch_proba)                                            # indice de la branch precise
            particle_branch = particle[index_rad][iDaughter][i_branch][index_subBranch]            # sampled particle emitted by the mother
            energy_branch =  e_branch[index_rad][iDaughter][i_branch][index_subBranch]             # energy of the particle emitted by the mother
            # probability_branch = p_branch[index_rad][iDaughter][i_branch][index_subBranch]         # probability of the sampled branch
            levelOftheDaughter = LevelDaughter[index_rad][iDaughter][i_branch][index_subBranch]    # Level of the daughter just after the particle emission from the mother
            level_before_trans = LevelDaughter[index_rad][iDaughter][i_branch][index_subBranch]
            if Display: print("\t Sampled decay branch:")
            if Display: print("\t\t Particle =               ", particle_branch)
            if Display: print("\t\t Energy of the particle = ", energy_branch, " keV")
            if Display: print("\t\t Level of the daughter nucleus = ", levelOftheDaughter)
            #========
            # Scoring
            #========
            e_sum = energy_branch                               # Update the Energy summary
            particle_vec.append(particle_branch)                # Update of the particle vector
            energy_vec.append(energy_branch)                    # Update of the energy of the particle
        else:
            transition_prob = tl.normalise(Transition_prob_sum[index_rad][iDaughter])
            index_transition_level = tl.sampling(transition_prob)
            levelOftheDaughter = levelNumber[index_rad][iDaughter][index_transition_level][0]
            if Display: print("\t Sampled decay branch:")
            if Display: print("\t\t Particle = isomeric transition, no particle")
            if Display: print("\t\t Level of the nucleus : ",levelOftheDaughter)
            e_sum = 0

        '''
        ==============
        I-1 Transition
        ==============
        '''  

        if Display: print("\t Subsequent isomeric transition")                       # finish with the mother / now with the daughter
        while levelOftheDaughter > 0:                                                # Go on the loop while the daughter nucleus is a its fundamental level (energy 0)
            i_level = levelNumber[index_rad][iDaughter].index([levelOftheDaughter])  # Find the position in the daughter level vector
            if transitionType[index_rad][iDaughter][i_level] != []:
                #====================================================================
                # Sampling of the transition in energy levels of the daughter nucleus
                #====================================================================
                probability_tran = tl.normalise(prob_trans[index_rad][iDaughter][i_level])   # normaliser la proba de transition 
                index_t = tl.sampling(probability_tran)                                      # indice de la transition
                if Display: print("\t\t Energy of the level = ", levelEnergy[index_rad][iDaughter][i_level], " keV")
                if Display: print("\t\t Transition type =          ", transitionType[index_rad][iDaughter][i_level][index_t])
                if Display: print("\t\t Energy of the transition = ", e_trans[index_rad][iDaughter][i_level][index_t], "keV")
                if Display: print("\t\t next level =               ", next_level[index_rad][iDaughter][i_level][index_t])
                #========
                # Scoring
                #========
                if transitionType[index_rad][iDaughter][i_level][index_t] == "GA":            # if it is a gamma that has been emitted
                    particle_vec.append("gamma")                                              # Update of the particle vector
                    energy_vec.append(e_trans[index_rad][iDaughter][i_level][index_t])        # Update the energy vector
                else:                                                                         # if not, it is a internal conversion, so an electron
                    particle_vec.append("electron")                                           # !!!!!!!!! it is OK for our model? Does the electron leave with the kinetic enegy of the transition 
                    energy_vec.append(e_trans[index_rad][iDaughter][i_level][index_t])        # Update the energy vector
                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EK":        # record that an electron is missing on the K shell of the dughter nucleus
                        particle_vec.append("Atom_K")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EL":       # record that an electron is missing on the L1 shell of the dughter nucleus
                        particle_vec.append("Atom_L")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EL1":       # record that an electron is missing on the L1 shell of the dughter nucleus
                        particle_vec.append("Atom_L1")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EL2":       # record that an electron is missing on the L2 shell of the dughter nucleus
                        particle_vec.append("Atom_L2")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EL3":       # record that an electron is missing on the L3 shell of the dughter nucleus
                        particle_vec.append("Atom_L3")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EM":        # record that an electron is missing on the M shell of the dughter nucleus
                        particle_vec.append("Atom_M")
                        energy_vec.append(0)

                    if transitionType[index_rad][iDaughter][i_level][index_t] == "EN":        # record that an electron is missing on the N shell of the dughter nucleus
                        particle_vec.append("Atom_N")
                        energy_vec.append(0)
                        
                e_sum += e_trans[index_rad][iDaughter][i_level][index_t]                      # Energy summary
                levelOftheDaughter = next_level[index_rad][iDaughter][i_level][index_t]       # set the next level
            else:
                i_level = levelNumber[index_rad][iDaughter].index([levelOftheDaughter])
                print("warning:pas de données de transition:daughter,niveau,niveau d'énergie",DaughterVec[index_rad][iDaughter],levelOftheDaughter,levelEnergy[index_rad][iDaughter][i_level] )
                levelOftheDaughter = 0   # set the next level
                
        if Display: print("\t Summary of the nuclear decay")
        if Display: print("\t\t particles : ", particle_vec)
        if Display: print("\t\t energy    : ", energy_vec, "keV")

        '''
        ==========================
        II. LA RELAXATION ATOMIQUE
        ==========================
        '''
        daughter_relax = DaughterVec[index_rad][iDaughter]
        for i_part in range(len(particle_vec)):
            relaxation = False
            if "Atom_K" in particle_vec[i_part] or "Atom_L" in particle_vec[i_part] or "Atom_M" in particle_vec[i_part]:
                relaxation = True
            while relaxation:
                tf,ef = tl.relaxation_atom(daughter_relax,Rad[index_rad],particle_vec[i_part])
                if tf == "XKA":
                    particle_vec[i_part] = "Atom_L"
                    particle_vec.append(tf)
                    energy_vec.append(ef)
                    relaxation = True
                elif tf == "XKB":
                    particle_vec[i_part] = "Atom_M"
                    particle_vec.append(tf)
                    energy_vec.append(ef)
                    relaxation = False
                elif tf == "XL":
                    particle_vec[i_part] = "Atom_M"
                    particle_vec.append(tf)
                    energy_vec.append(ef)
                    relaxation = False
                elif tf == "Auger K":
                    particle_vec[i_part] = "Atom_L"
                    particle_vec.append(tf)
                    energy_vec.append(ef)
                    relaxation = True
                elif tf == "Auger L":
                    particle_vec[i_part] = "Atom_M"
                    particle_vec.append(tf)
                    energy_vec.append(ef)
                    relaxation = False
                else:
                    if Display: print("untermined x or Auger")
                    relaxation = False
                e_sum += ef
        if Display: print("\t Summary of the atomic relaxation")
        if Display: print("\t\t particles : ", particle_vec)            
        if Display: print("\t\t energy    : ", energy_vec, "keV")

        '''
        ==========================================================
        III. INTERACTION RAYONNEMENT/MATIERE + SPECTRES D'EMISSION
        ==========================================================
        '''
        for i, p in enumerate(particle_vec):
            if p == "beta":
                e_b,p_b = tl.readBetaShape(rad_i,"beta-",level_before_trans)   # read the data of BetaShape
                index_beta_energy = tl.sampling(p_b)                           # sampling energy of beta
                particle_vec[i] = "electron"
                energy_vec[i] = e_b[index_beta_energy]
                energy_vec[i] = tl.energie_dep_beta(energy_vec[i])

            if p == "beta+":
                e_b,p_b = tl.readBetaShape(rad_i,"beta+",level_before_trans)
                index_beta_energy = tl.sampling(p_b)
                particle_vec[i] = "positron"
                energy_vec[i] = e_b[index_beta_energy]
                energy_vec[i] = tl.energie_dep_beta(energy_vec[i]) # treated as beta-
                particle_vec.append("gamma")
                particle_vec.append("gamma")
                energy_vec.append(511)
                energy_vec.append(511)

            if p == "gamma" or p == "XKA" or p == "XKB" or p == "XL":
                energy_vec[i] = tl.energie_dep_gamma(energy_vec[i],v=10)          # sampling energy free from photon
                particle_vec[i] = "electron"
            if p == "Auger K" or p == "Auger L":
                particle_vec[i] = "electron"
                energy_vec[i] = tl.energie_dep_beta(energy_vec[i])
        if Display: print("\t Summary of the final charged particles")
        if Display: print("\t\t particles : ", particle_vec)
        if Display: print("\t\t energy    : ", energy_vec, "keV")

        '''
        ====================
        IV. LA SCINTILLATION
        Calculation of the scintillation quenching with the Birks Model
        ====================
        '''
        e_quenching=[]
        for i, p in enumerate(particle_vec):
            e_discrete = np.linspace(0,energy_vec[i],nE) # vector for the quenched  energy calculation keV
            delta_e = e_discrete[2]-e_discrete[1]  #keV
            if p == "alpha":
                energy_vec[i] = np.cumsum(delta_e/(1+kB*tl.stoppingpowerA(e_discrete)))[-1]
                e_quenching.append(energy_vec[i])
            elif p == "electron" or p == "positron":
                energy_vec[i] = tl.E_quench_e(energy_vec[i]*1e3,kB*1e3,nE)*1e-3
                e_quenching.append(energy_vec[i])
            else:
                e_quenching.append(0)
        if Display: print("\t Summary of the estimation of quenched energies")
        
        if Display: print("\t\t energy_vec      : ", energy_vec, "keV")
        if Display: print("\t\t quenched energy : ", e_quenching, "keV")

        '''
        ====================
        V. LE MESURE TDCR
        ====================
        '''
        
        if mode2=="sym":
            p_nosingle = np.exp(-L*np.sum(np.asarray(e_quenching))/3) # probability to have 0 electrons in a PMT
            p_single = 1-p_nosingle                                    # probability to have at least 1 electrons in a PMT
            efficiency_S.append(p_single)
            efficiency_T.append(p_single**3)
            efficiency_D.append(3*(p_single)**2-2*efficiency_T[-1])
            if Display: print("\t Summary of TDCR measurement")
            if Display: print("\t\t Efficiency of single events: ", efficiency_S[-1])
            if Display: print("\t\t Efficiency of double events: ", efficiency_D[-1])
            if Display: print("\t\t Efficiency of triple events: ", efficiency_D[-1])
        elif mode2=="asym":
            pA_nosingle = np.exp(-L[0]*np.sum(np.asarray(e_quenching))/3) # probability to have 0 electrons in a PMT
            pA_single = 1-pA_nosingle                                    # probability to have at least 1 electrons in a PMT
            pB_nosingle = np.exp(-L[1]*np.sum(np.asarray(e_quenching))/3) # probability to have 0 electrons in a PMT
            pB_single = 1-pB_nosingle                                    # probability to have at least 1 electrons in a PMT
            pC_nosingle = np.exp(-L[2]*np.sum(np.asarray(e_quenching))/3) # probability to have 0 electrons in a PMT
            pC_single = 1-pC_nosingle                                    # probability to have at least 1 electrons in a PMT
            
            efficiency_AB.append(pA_single*pB_single)
            efficiency_BC.append(pB_single*pC_single)
            efficiency_AC.append(pA_single*pC_single)
            efficiency_T.append(pA_single*pB_single*pC_single)
            
            efficiency_D.append(efficiency_AB[-1]+efficiency_BC[-1]+efficiency_AC[-1]-2*efficiency_T[-1])
            efficiency_S.append(pA_single+pB_single+pC_single-efficiency_D[-1]-efficiency_T[-1])
            
            

    '''
    ====================
    VI. CALCULATION OF THE FINAL ESTIMATORS
    ====================
    '''
    if mode2=="sym":
        mean_efficiency_T = np.mean(efficiency_T) # average
        std_efficiency_T = np.std(efficiency_T)/np.sqrt(N)   # standard deviation
        mean_efficiency_D = np.mean(efficiency_D)
        std_efficiency_D = np.std(efficiency_D)/np.sqrt(N)
        mean_efficiency_S = np.mean(efficiency_S)
        std_efficiency_S = np.std(efficiency_S)/np.sqrt(N)
        TDCR_calcul = mean_efficiency_T/mean_efficiency_D
    elif mode2=="asym":
        mean_efficiency_T = np.mean(efficiency_T) # average
        std_efficiency_T = np.std(efficiency_T)/np.sqrt(N)   # standard deviation
        mean_efficiency_D = np.mean(efficiency_D)
        std_efficiency_D = np.std(efficiency_D)/np.sqrt(N)
        mean_efficiency_S = np.mean(efficiency_S)
        std_efficiency_S = np.std(efficiency_S)/np.sqrt(N)
        mean_efficiency_AB = np.mean(efficiency_AB)
        std_efficiency_AB = np.std(efficiency_AB)/np.sqrt(N)
        mean_efficiency_BC = np.mean(efficiency_BC)
        std_efficiency_BC = np.std(efficiency_BC)/np.sqrt(N)
        mean_efficiency_AC = np.mean(efficiency_AC)
        std_efficiency_AC = np.std(efficiency_AC)/np.sqrt(N)
        TDCR_calcul = mean_efficiency_T/mean_efficiency_D
        TABmodel = mean_efficiency_T/mean_efficiency_AB
        TBCmodel = mean_efficiency_T/mean_efficiency_BC
        TACmodel = mean_efficiency_T/mean_efficiency_AC
        
        
    #    x = np.arange(np.mean(efficiency_T),1.001,0.001)
    #    plt.figure("efficiency distribution")
    #    plt.clf()
    #    plt.hist(np.asarray(efficiency_D),bins=x,label="Efficiency of double coincidences")[0]
    #    plt.hist(np.asarray(efficiency_T),bins=x,label="Efficiency of triple coincidences")[0]
    #    plt.xlabel("Efficiency", fontsize = 14)
    #    plt.ylabel(r"Number of counts", fontsize = 14)
    #    plt.legend(fontsize = 12)
    #    plt.savefig('Effdistribution.png')
    
    #    x = np.arange(np.mean(tdcr),1.001,0.001)
    #    plt.figure("TDCR distribution")
    #    plt.clf()
    #    plt.hist(np.asarray(tdcr),bins=x,label="Calculated TDCR")[0]
    #    plt.plot(x,st.norm.pdf(x, TDCR_measure, u_TDCR_measure),label="measured TDCR")[0]
    #    plt.xlabel("Efficiency", fontsize = 14)
    #    plt.ylabel(r"Number of counts", fontsize = 14)
    #    plt.legend(fontsize = 12)
    #    plt.savefig('TDCRdistribution.png')
    
    if mode2=="sym":
        res=(TDCR_calcul-TD)**2
    elif mode2=="asym":
        res=(TAB-TABmodel)**2+(TBC-TBCmodel)**2+(TAC-TACmodel)**2
    
    if mode == "res":
        return res
    if mode == "eff":
        return mean_efficiency_S, std_efficiency_S, mean_efficiency_D, std_efficiency_D, mean_efficiency_T, std_efficiency_T


"""
This program aims to calculate the relative composition of radionuclides
decaying in chain using the Bateman model implemented in the radioactivedecay module form (MIT)

Romain Coulon, Jialin Hu
Bureau International des Poids et Mesures
"""

# pip install radioactivedecay
import radioactivedecay as rd
import matplotlib.pyplot as plt

rad = "Ra-223" # Radionuclide
A0 = 1         # Inital activity
unit = 'Bq'    # unit for the activity
decaytime = 2  # decay time
unitt = 'd'    # unit for the time 

rad_t0 = rd.Inventory({rad: A0}, unit)
rad_t1 = rad_t0.decay(decaytime, unitt)
rad_t1.activities(unit)
print(rad_t1)

decayGraph = rad_t0.plot(decaytime, unitt, yunits=unit)
plt.savefig('decayGraph.png')

nuc = rd.Nuclide(rad)
decayChain = nuc.plot()
plt.savefig('decayChain.png')



# -*- coding: utf-8 -*-
"""
Created on Wed Jul  5 10:04:23 2023

@author: romain.coulon
"""
import TDCR_model_lib as tl
import TDCRPy as td
import matplotlib.pyplot as plt

## INPUT OF THE MODEL
# N=1                   # number of simulated decay (MC trials)
N= 10000
Rad="Co-60, H-3"            # list of radionuclides (Na-24)
# Rad = ["Cs-137"]
pmf_1="0.5, 0.5"                # relative abondance (pmf)
kB =[1.0e-5]
# kB = [0.8e-5, 0.9e-5, 1.0e-5, 1.1e-5, 1.2e-5]    # Birks constant in cm/keV
L=[1.13]
# L = np.logspace(-3,1,50) # Free paramete in keV-1

Record = False                  # to record the efficiency curves
# Record = True                  # to record the efficiency curves
# Display = False               # to display calculation results on the console
Display = True                # to display calculation results on the console
# RHO = 0.96         #density of absorber (Toluene) g/cm3
RHO = 0.98           #density of absorber (UG + H20) g/cm3
nE = 1000            #number of bin to discretize the energy vector for scintillation quenching calculation
mode = "eff"
#meta_vec = ["Am-242m","Pa-234m","Pm-148m","Pr-144m","Xe-133m","Te-127m","Ag-110m","Ag-108m","Tc-99m","Nb-95m","Y-90m","Mn-52m"]

TDCR_measure = 0.977784        # Measured TDCR value
u_TDCR_measure = 0.000711      # standard uncertainty
TAB=0.8
TBC=0.8
TAC=0.8

for kB_i in kB: # Loop on the kB
    mean_efficiency_S = []  # efficiency of single counte rate
    std_efficiency_S = []   # std
    mean_efficiency_T = []  # efficiency of triple coincidence counte rate
    std_efficiency_T = []   # std
    mean_efficiency_D = []  # efficiency of double coincidence counte rate
    std_efficiency_D = []   # std
    TDCR_calcul = []
    for L_i in L: # loop on the free parameter values
        efficiency_S = []        # results calculated efficiency for single events
        efficiency_D = []        # results calculated efficiency for double coincidence
        efficiency_T = []        # results calculated efficiency for triple coincidence
        
        out=td.TDCRPy(L_i,TDCR_measure,TAB, TBC, TAC, Rad,pmf_1,10,kB_i,RHO,nE,mode="eff",mode2="sym",Display=True)
        efficiency_S.append(out[0])
        std_efficiency_S.append(out[1])
        efficiency_D.append(out[2])
        std_efficiency_D.append(out[3])
        efficiency_T.append(out[4])
        std_efficiency_T.append(out[5])
        
        if len(L) > 1: print("\n\t\t Progress = ", round(100*(L.tolist().index(L_i)+1)/len(L), 1), " %")
        print("\t\tradionuclide(s): ", Rad)
        print("\t\t TDCR calculation _ kB = ", kB_i, "cm/keV")
        print("\t\t Free parameter = ", L_i, " keV-1")
        print("\t\t Efficiency of Triple coincident events = ", round(100*efficiency_T[-1],3), "+/-", round(100*std_efficiency_T[-1],3), " %")
        print("\t\t Efficiency of Double coincident events = ", round(100*efficiency_D[-1],3), "+/-", round(100*std_efficiency_D[-1],3), " %")
        print("\t\t Efficiency of Single events = ", round(100*efficiency_S[-1],3), "+/-", round(100*std_efficiency_S[-1],3), " %")

    if len(mean_efficiency_S)>1:
        plt.figure("Efficiency curve I")
        plt.clf()
        plt.title(''.join(Rad))
        plt.errorbar(L, mean_efficiency_S, yerr = std_efficiency_S, fmt=".b", label = "S")
        plt.errorbar(L, mean_efficiency_D, yerr = std_efficiency_D, fmt=".k", label = "D")
        plt.errorbar(L, mean_efficiency_T, yerr = std_efficiency_T, fmt=".r", label = "T")
        plt.xscale("log")
        plt.xlabel(r"$L$ /keV$^{-1}$", fontsize = 14)
        plt.ylabel(r"$\epsilon$", fontsize = 14)
        plt.legend(fontsize = 12)
        if Record: plt.savefig("EfficiencyCurves/"+''.join(Rad)+"/fom_"+''.join(Rad)+"_"+str(kB_i)+".png")
        plt.close()
                
        plt.figure("Efficiency curve II")
        plt.clf()
        plt.title(''.join(Rad))
        plt.errorbar(TDCR_calcul, mean_efficiency_S, yerr=std_efficiency_S, fmt=".b", label = "S")
        plt.errorbar(TDCR_calcul, mean_efficiency_D, yerr=std_efficiency_D, fmt=".k", label = "D")
        plt.errorbar(TDCR_calcul, mean_efficiency_T, yerr=std_efficiency_T, fmt=".r", label = "T")
        plt.xlabel(r"$\epsilon_T/\epsilon_D$", fontsize = 14)
        plt.ylabel(r"$\epsilon_D$", fontsize = 14)
        plt.legend(fontsize = 12)
        if Record: plt.savefig("EfficiencyCurves/"+''.join(Rad)+"/tdcr_"+''.join(Rad)+"_"+str(kB_i)+".png")
        plt.close()
    
    if Record:
        tl.writeEffcurves(L, mean_efficiency_S, std_efficiency_S, Rad, pmf_1, kB_i, "S")
        tl.writeEffcurves(L, mean_efficiency_D, std_efficiency_D, Rad, pmf_1, kB_i, "D")
        tl.writeEffcurves(L, mean_efficiency_T, std_efficiency_T, Rad, pmf_1, kB_i, "T")


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 14 20:43:26 2023

@author: romain.coulon
"""
import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as sg
import scipy.optimize as opt
import sys
sys.path.insert(1, 'G:\Python_modules\BIPM_RI_PyModules')
import  TDCRcalculation as dtc  
    
    
def readEff(Rad, kB, SDT):
    file = open("G:\Python_modules\TDCRPy\TDCRPy\Code\\EfficiencyCurves/"+''.join(Rad)+"/Eff"+SDT+"_"+''.join(Rad)+'_[1]_'+str(kB)+".txt","r")
    L=[]
    p=[]
    up=[]
    for row in file:
        H=row.split(" ")
        L.append(float(H[0]))
        p.append(float(H[1]))
        up.append(float(H[2]))
    return L, p, up

def readProfil(Rad,kB,SDT):
    Lv, pSv, upSv = readEff(Rad, kB, SDT)
    pSv = sg.savgol_filter(pSv, 7, 2)
    return Lv, np.asarray(pSv), upSv

def res(L, TD, Lv, pDv, pTv):
    return (np.interp(L, Lv, pTv)/np.interp(L, Lv, pDv)-TD)**2

def effTDCR(TD, Rad, kB):
    kBv = [0.8e-5, 0.9e-5, 1.0e-5, 1.1e-5, 1.2e-5]
    Lv0 = readProfil(Rad,0.8e-5,"D")[0]
    pDv=[]; pTv=[]; Lv1=[]; effD1=[]; effT1=[]
    for i in kBv:
        pDv.append(readProfil(Rad,i,"D")[1])
        pTv.append(readProfil(Rad,i,"T")[1])
        Lv1.append(opt.minimize_scalar(res, args=(TD, Lv0, pDv[-1], pTv[-1]), method='golden').x)
        effD1.append(np.interp(Lv1[-1], Lv0, pDv[-1]))
        effT1.append(np.interp(Lv1[-1], Lv0, pTv[-1]))
    L = np.interp(kB, kBv, Lv1)
    effD = np.interp(kB, kBv, effD1)
    effT = np.interp(kB, kBv, effT1)
     
    # plt.figure("eff_D = f(kB)")
    # plt.clf()
    # plt.title('  ')
    # plt.plot(kBv, effD1,'-b')
    # plt.plot(kB, effD, 'or')
    # plt.xlabel(r"$kB$ /(cm keV$^{-1}$)", fontsize = 14)
    # plt.ylabel(r"$\epsilon_D$", fontsize = 14)
    # plt.legend(fontsize = 12)
    # # plt.close()
    
    # plt.figure("eff_T = f(kB)")
    # plt.clf()
    # plt.title('  ')
    # plt.plot(kBv, effT1,'-b')
    # plt.plot(kB, effT, 'or')
    # plt.xlabel(r"$kB$ /(cm keV$^{-1}$)", fontsize = 14)
    # plt.ylabel(r"$\epsilon_T$", fontsize = 14)
    # plt.legend(fontsize = 12)
    # # plt.close()
    
    # plt.figure("L = f(kB)")
    # plt.clf()
    # plt.title('  ')
    # plt.plot(kBv, Lv1,'-b')
    # plt.plot(kB, L, 'or')
    # plt.xlabel(r"$kB$ /(cm keV$^{-1}$)", fontsize = 14)
    # plt.ylabel(r"$L$ /keV$^{-1}$", fontsize = 14)
    # plt.legend(fontsize = 12)
    # # plt.close()
    
    return L, effD, effT

def plotEffProfil(Rad,kB):
    kBv = [0.8e-5, 0.9e-5, 1.0e-5, 1.1e-5, 1.2e-5]
    Lv0 = readProfil(Rad,0.8e-5,"D")[0]
    pDv=[]; pTv=[]; pSv=[]
    for i in kBv:
        pDv.append(readProfil(Rad,i,"D")[1])
        pTv.append(readProfil(Rad,i,"T")[1])
        pSv.append(readProfil(Rad,i,"S")[1])
    effS=[]; effD=[]; effT=[]
    print("S",np.asarray(pSv)[:,0])
    for i in range(len(Lv0)):
        effS.append(np.interp(kB, kBv, np.asarray(pSv)[:,i]))
        effD.append(np.interp(kB, kBv, np.asarray(pDv)[:,i]))
        effT.append(np.interp(kB, kBv, np.asarray(pTv)[:,i]))
    tdcr = np.asarray(effT)/np.asarray(effD)
    
    plt.figure("Efficiency curve eff = f(L)")
    plt.clf()
    plt.title('  ')
    plt.plot(Lv0, effS,'-b', label = r"$\epsilon_{S}$, $kB$ = "+str(kB)+" cm/keV")
    plt.plot(Lv0, effD,'-k', label = r"$\epsilon_{D}$, $kB$ = "+str(kB)+" cm/keV")
    plt.plot(Lv0, effT,'-r', label = r"$\epsilon_{T}$, $kB$ = "+str(kB)+" cm/keV")
    plt.xscale("log")
    plt.xlabel(r"$L$ /keV$^{-1}$", fontsize = 14)
    plt.ylabel(r"$\epsilon$", fontsize = 14)
    plt.legend(fontsize = 12)
    # plt.close()
    
    plt.figure("Efficiency curve eff = f(TDCR)")
    plt.clf()
    plt.title('  ')
    plt.plot(tdcr, effS,'-b', label = r"$\epsilon_{S}$, $kB$ = "+str(kB)+" cm/keV")
    plt.plot(tdcr, effD,'-k', label = r"$\epsilon_{D}$, $kB$ = "+str(kB)+" cm/keV")
    plt.plot(tdcr, effT,'-r', label = r"$\epsilon_{T}$, $kB$ = "+str(kB)+" cm/keV")
    # plt.xscale("log")
    plt.xlabel(r"$\epsilon_T/\epsilon_D$", fontsize = 14)
    plt.ylabel(r"$\epsilon$", fontsize = 14)
    # bbox_inches = "tight", format = "png", dpi = 500
    plt.legend(fontsize = 12)
    # plt.close()
    
    
def plotEffProfilkB(Rad,SDT): # plot the fitted efficiecny curves for a range of kB 
    kBv = [0.8e-5, 0.9e-5, 1.0e-5, 1.1e-5, 1.2e-5]   # kB vector
    plt.figure("Efficiency curve eff = f(TDCR)")
    plt.clf()
    plt.title('  ')
    for i in kBv:                                    # loop in the kB vector
        pDv = readProfil(Rad,i,"D")[1]               # probability vector of events
        pTv = readProfil(Rad,i,"T")[1]               # probability vector of triple coincidence events
        pSv = readProfil(Rad,i,"S")[1]               # probability vector of single events
        tdcr = np.asarray(pTv)/np.asarray(pDv)       # tdcr vector
        if SDT == "S":
            plt.plot(tdcr, pSv, label = r"$kB$ = "+str(i)+" cm/keV")
            plt.ylabel(r"$\epsilon_S$", fontsize = 14)
        if SDT == "D":
            plt.plot(tdcr, pDv, label = r"$kB$ = "+str(i)+" cm/keV")
            plt.ylabel(r"$\epsilon_D$", fontsize = 14)
        if SDT == "T":
            plt.plot(tdcr, pDv, label = r"$kB$ = "+str(i)+" cm/keV")
            plt.ylabel(r"$\epsilon$", fontsize = 14)
        # plt.xscale("log")
        plt.xlabel(r"$\epsilon_T/\epsilon_D$", fontsize = 14)
        # bbox_inches = "tight", format = "png", dpi = 500
        plt.legend(fontsize = 12)
        plt.savefig("EfficiencyCurves/"+Rad+"/tdcr_"+Rad+".png")
        # plt.close()

def plotSmoothing(Rad,SDT): # plot the fitted efficiecny curves for a range of kB 
    kB = 1.0e-5
    plt.figure(r"Fitting of calculated points - $kB$ = "+str(kB)+" cm/keV")
    plt.clf()
    plt.title('  ')
    pDv = readProfil(Rad,kB,"D")[1]               # probability vector of events
    pts_D = readEff(Rad, kB, "D")
    pTv = readProfil(Rad,kB,"T")[1]               # probability vector of triple coincidence events
    pts_T = readEff(Rad, kB, "T")
    pSv = readProfil(Rad,kB,"S")[1]               # probability vector of single events
    pts_S = readEff(Rad, kB, "S")
    tdcr = np.asarray(pTv)/np.asarray(pDv)       # tdcr vector fit
    tdcr_x = np.asarray(pts_T[1])/np.asarray(pts_D[1]) # calculated points
    if SDT == "S":
        plt.plot(tdcr_x, pts_S[1], "ok", label = r"calculated points")
        plt.plot(tdcr, pSv, label = r"fitted function")
        plt.ylabel(r"$\epsilon_S$", fontsize = 14)
    if SDT == "D":
        plt.plot(tdcr_x, pts_D[1], "ok", label = r"calculated points")
        plt.plot(tdcr, pDv, label = r"fitted function")
        plt.ylabel(r"$\epsilon_D$", fontsize = 14)
    if SDT == "T":
        plt.plot(tdcr_x, pts_T[1], "ok", label = r"calculated points")
        plt.plot(tdcr, pDv, label = r"fitted function")
        plt.ylabel(r"$\epsilon$", fontsize = 14)
    # plt.xscale("log")
    plt.xlabel(r"$\epsilon_T/\epsilon_D$", fontsize = 14)
    # bbox_inches = "tight", format = "png", dpi = 500
    plt.legend(fontsize = 12)
    plt.savefig("EfficiencyCurves/"+Rad+"/tdcr_"+Rad+"_fit.png")
    # plt.close()

def plotL(radv,kB,L): # plot the fitted efficiecny curves for a range of kB 
    
    effD = []; effT = []
    for Rad in radv:
        out = readProfil(Rad,kB,"D")               # probability vector of double events
        effD.append(np.interp(L, out[0], np.asarray(out[1])))
        out = readProfil(Rad,kB,"T")               # probability vector of triple events
        effT.append(np.interp(L, out[0], np.asarray(out[1])))
    
    effD = np.asarray(effD)
    effT = np.asarray(effT)
    
    x = effT/effD
    plt.figure(r"Fitting of calculated points")#" - $kB$ = "+str(kB)+" cm/keV - $L$ = "+str(L)+" keV$^{-1}$")
    plt.clf()
    plt.title('  ') 
    plt.plot(x, effD, "ok", label="TDCR model")
    for ix, c in enumerate(x):
        plt.text(c, effD[ix]+0.000, radv[ix], color="b")
    plt.errorbar(0.9788, 0.9767, xerr=0.0012, yerr=0.011, fmt="or", label="experiemental result")
    plt.text(0.9788, 0.9767, "Co-60", color = "r")
    plt.ylabel(r"$\epsilon_D$", fontsize = 14)
    plt.xlabel(r"$\epsilon_T/\epsilon_D$", fontsize = 14)
    # plt.xscale("log")
    plt.legend(fontsize = 12)
    plt.savefig("EfficiencyCurves/ESIRIC/EfficiencyCurve.png")
    plt.show()
    
    # plt.xscale("log")

    # bbox_inches = "tight", format = "png", dpi = 500

    # plt.close()

## PLOT EFFICIENCY CURVES FOR SEVERAL KB VALUES
# plotEffProfilkB("C-14","D")

## PLOT EFFICIENCY CURVE A GIVEN KB VALUE
# plotEffProfil("Fe-55", 1.05E-5)

## PLOT THE FIT OF THE EFFICIENCY FUNCTION
# plotSmoothing("C-14","D")

## PLOT THE EFFICIENCY FOR A GIVEN FREE PARAMETER
# plotSmoothing("C-14","D")

## CALCULATION OF THE EFFICIENCY
# Reference Point Co-60 => TDCR=0.9788 Eff_D = 0.9767 => L = 

# radv = ["H-3", "He-6", "Be-7", "C-11", "C-14", "N-13", "O-15", "F-18", "Na-22", \
# "Na-24", "Al-26", "P-32", "P-33", "S-35", "Cl-36", "Ar-41", "K-40", \
# "Ca-41", "Ca-45", "Sc-44", "Sc-46", "Sc-47", "Ti-44", "Cr-51", \
# "Mn-52", "Mn-52m", "Mn-54", "Mn-56", "Fe-52", "Fe-55", "Fe-59", "Co-56", \
# "Co-57", "Co-58", "Co-60", "Ga-66", "Ni-57", "Ni-59", "Ni-63", "Zn-65", \
# "I-123", "Sb-124", "Sb-125", "Te-123m", "Te-127", "Te-127m", "Te-132", \
# "Pu-241"]


radv = ["Fe-55"]
kB = 1.0e-5
L = 1.13
plotL(radv, kB, L)

# AB = 657.296
# BC = 695.919
# AC = 693.579
# T = 448.868
# D = AB+BC+AC-2*T
# out = effTDCR(T/D, "H-3", 1.05E-5) # H-3
# plotEffProfil("H-3", 1.05E-5)

# AB = 864.95
# BC = 1006.58
# AC = 968.45
# T = 700
# D = 1000
# XX = (D + 2*T)/3
# print(XX)
# print(T/D)
# D = AB+BC+AC-2*T

# TDCR = [0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
# kB = 0.8E-5
# rad = "C-14"
# TDCR.reverse()
# for t in TDCR:
#     eff = effTDCR(t, rad, kB)[1] # H-3
#     eff2 = dtc.I2calc(t, 0, 0, 0, rad, kB*1e3)[2]
#     digRound = 2
#     print("Efficiency of double coincidences = ", round(100*eff2, digRound),"% - ", round(100*eff, digRound),"%")

    



# DT = 0.977784348 # Co-60 NIST
# uDT = 0.000711023 # Co-60 NIST
# D = 84679.21179
# nMc = 1e3
# eff_v = []
# kB1 = 0.8e-5
# kB2 = 1.2e-5
# for i in range(int(nMc)):
#     dti = np.random.normal(DT,uDT)
#     kBi = np.random.uniform(kB1,kB2)
#     eff_v.append(effTDCR(dti, "Co-60", kBi)[1]) # Co-60 NIST
# eff = np.median(eff_v)
# u_eff = np.std(eff_v)
# A = D/eff
# uA = D*u_eff/eff**2
# result 87099      67             (err = 0.0002) kB = 0.01
# result 87045      88             kB = [0.008 - 0.012]
# I2 86640.31523	14.64334414
# A_NIST = 86920    200

# out = effTDCR(657.296, 695.919, 693.579, 448.868) # Co-60
# print("Efficiency of double coincidences = ", round(100*eff, digRound),"+/-", round(100*u_eff, digRound),"%")
# print("Efficiency of triple coincidences = ", round(100*out[2], digRound),"%")
# print("Activity from double coincidences = ", round(A, digRound), "+/-", round(uA, digRound) ,"Bq")
# print("Activity of triple coincidences = ", round(T/out[2], digRound), "Bq")
# print("Efficiency of double coincidences (asym) = ", round(100*out[3], digRound),"%")
# print("Efficiency of triple coincidences (asym) = ", round(100*out[5], digRound),"%")
# print("Activity from double coincidences (asym) = ", round(D/out[3], digRound), "Bq")
# print("Activity of triple coincidences (asym) = ", round(T/out[5], digRound), "Bq")


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 13 19:15:23 2023

@author: romain.coulon
"""
import TDCRPy

TDCRPy.TDCRPy(1, 0.5, 0.5, 0.5, 0.5, "H-3", "1", 10, 0.00001, 1,10, "eff", "sym", Display=True)


TDCRPy.TDCRPy(L, TD, TAB, TBC, TAC, Rad, pmf_1, N, kB, RHO, nE, mode, mode2)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 23 16:04:46 2023

Library of function of the TDCRpy code

@author: Romain Coulon, Jialin Hu
Bureau International des Poids et Mesures
"""

### IMPORT Python Module
import urllib.request as rq
import numpy as np
from numpy.core.multiarray import where
import zipfile as zf
# import statsmodels.api as sm
import matplotlib.pyplot as plt
import time
import re
absolutePath = False


def TicTocGenerator():
    """
    Generator that returns time differences
    """
    ti = 0           # initial time
    tf = time.time() # final time
    while True:
        ti = tf
        tf = time.time()
        yield tf-ti # returns the time difference

TicToc = TicTocGenerator() # create an instance of the TicTocGen generator

# This will be the main function through which we define both tic() and toc()
def toc(tempBool=True):
    """
    Prints the time difference yielded by generator instance TicToc
    """
    
    tempTimeInterval = next(TicToc)
    if tempBool:
        print( "Elapsed time: %f seconds.\n" %tempTimeInterval )

def tic():
    """
    Records a time in TicToc, marks the beginning of a time interval
    """
    toc(False)

def normalise(p_x):
    p_array = np.array(p_x)
    if len(p_x)>1:
        p_somme = sum(p_array)
        if p_somme>0.0:
            p_array = p_array/p_somme
    else:
        p_somme = p_x[0]
        p_array = p_array/p_somme
    p = list(p_array)
    return p
#print(normalise([1, 0.9974641, 0]))

def sampling(p_x):
    """
    This function aims to sample in a pdf or a pmf

    Parameters
    ----------
    p_x : float vector
        Probability Density (or mass) Function (PDF or PMF) of the random variable x.


    Returns
    -------
    i : integer
        index in x pointing the sampled value of the random variable X.
    """

    cf = np.cumsum(p_x) # Cummulative Density (or mass) Function (CDF or CMF)
    trial = float(np.random.rand(1)) # trial ~ U(0,1)
    
    for i, p in enumerate(cf):
        if p> trial: break
    return i


def readPenNuc(rad):
    """
    This function reads the PenNuc files published in the DDEP web page
    http://www.lnhb.fr/donnees-nucleaires/donnees-nucleaires-tableau/
    
    Parameters
    ----------
    rad : string
        Indentifier of the radionuclide (e.g. Na-22).

    Returns
    -------
    out : list
        list of elements in list describing the decay of the radionuclide with,
        *particle* is the list of particle of the decay branches 
        *p_branch* is the list of probabilty of the decay branches
        *e_branch* is the list of energy of the decay branches
        *LevelDaughter* is the list of energy levels of the daughter nucleus following the decay branch
        *levelNumber* is the list of number of the energy level of the daughter nucleus following the decay branch
        *prob* is the list of probabilty of isomeric transitions
        *levelEnergy* is the energy levels of the isomeric transitions
        *transitionType* is the type of isomeric transitions
        *e_trans* is the energy of the isomeric transitions
        *next_level* is the possible next energy level following a given isomeric transtion 
        *Q_value_vec[dd]* is the Q-value of the decay when decaying to the daughter of index dd
        *Daughter_vec[dd]* is the vecteur of daugher nuclei
        *Pdaughter_vec[dd]* is the probability of daughter nuclei
    """
    
    url = "http://www.lnhb.fr/nuclides/"+rad+".PenNuc.txt"
    file = rq.urlopen(url)
    
    # Format the data 
    decayData = []
    for line in file:
      decayData.append(line.decode("utf-8"))
      if "NDA " in decayData[-1]:    decayData[-1] = decayData[-1].replace("NDA ", "NDA; ") # number of daughter
      if "DAU " in decayData[-1]:    decayData[-1] = decayData[-1].replace("DAU ", "DAU; ") # daughter
      if "DDE " in decayData[-1]:    decayData[-1] = decayData[-1].replace("DDE ", "DDE; ") # daughter description : probability of disintegration to NDA, uncertainty, number of excited levels of NDA; number of branche to NDA
      if "Q " in decayData[-1]:      decayData[-1] = decayData[-1].replace("Q ", "Q; ")     # total energy of the branch, uncertainty 
      if "ALP " in decayData[-1]:    decayData[-1] = decayData[-1].replace("ALP ", "ALP; ") # type of the disintegration = alpha
      if "CK " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CK ", "CK; ") # type of the disintegration = Electron Capture K shell
      if "CL " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL ", "CL; ") # type of the disintegration = Electron Capture L shell 1
      if "CL1 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL1 ", "CL1; ") # type of the disintegration = Electron Capture L shell 1
      if "CL2 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL2 ", "CL2; ") # type of the disintegration = Electron Capture L shell 2
      if "CM " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CM ", "CM; ") # type of the disintegration = Electron Capture M shell
      if "CN " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CN ", "CN; ") # type of the disintegration = Electron Capture N shell
      if "BEM " in decayData[-1]:    decayData[-1] = decayData[-1].replace("BEM ", "BEM; ") # type disintegration-beta- :branching ratio, uncertainty, level fed in daughter, energy, uncertainty,prohibition factor for beta emission
      if "BEP " in decayData[-1]:    decayData[-1] = decayData[-1].replace("BEP ", "BEP; ") # type disintegration-beta+ :branching ratio, uncertainty, level fed in daughter, energy, uncertainty,prohibition factor for beta emission
      if "LED " in decayData[-1]:    decayData[-1] = decayData[-1].replace("LED ", "LED; ") # level description:energy of level,uncertainty, number of transitions that depopulate this level,level time,uncertainty,level number
      if "GA " in decayData[-1]:     decayData[-1] = decayData[-1].replace("GA ", "GA; ")   # type transition: gamma
      if "EK " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EK ", "EK; ")   # type transition: internal conversion -- electron K
      if "EL1 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL1 ", "EL1; ") # type transition: internal conversion -- electron L1
      if "EL2 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL2 ", "EL2; ") # type transition: internal conversion -- electron L2
      if "EL3 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL3 ", "EL3; ") # type transition: internal conversion -- electron L3
      if "EM " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EM ", "EM; ")   # type transition: internal conversion -- electron M
      if "EN " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EN ", "EN; ")   # type transition: internal conversion -- electron N
      if "COM" in decayData[-1]:     decayData[-1] = decayData[-1].replace("COM ", "COM; ")   # commentaire
      decayData[-1]=decayData[-1].split(";")
      if "\r\n" in decayData[-1][-1]:  decayData[-1][-1] = decayData[-1][-1].replace("\r\n", " ")
    
   
      
    """
    LOOP IN NDAUGH (Daughters)
    """
    Daughter_vec = []; Pdaughter_vec = []; Q_value_vec = []
    for d in decayData:
      if d[0] == "NDA": Ndaughter = int(d[1])   # Read the number of daughter
      if d[0] == "DAU" :
          d[1] = d[1].replace(" ","")
          Daughter_vec.append(d[1])        # Read the daughter
      if d[0] == "DDE": Pdaughter_vec.append(float(d[1])) # Read the probability of disintegration to daughter
      if d[0] == "Q":   Q_value_vec.append(float(d[1]))   # Read the Q-value
    
    out = []
    for dd in range(Ndaughter):
        
        # index_Daughter = sampling(np.asarray(Pdaughter_vec)/sum(np.asarray(Pdaughter_vec)))
        # Daughter = Daughter_vec[index_Daughter]
        # Pdaughter =  Pdaughter_vec[index_Daughter]
        # Q_value = Q_value_vec[index_Daughter]    
         
        
        """
        LOOP IN NBRANCH (Branches of each daughter)
        """
        # Probablity vector of the decay branch
        particle=[]; LevelDaughter = []; p_branch = []; u_p_branch = []; e_branch = []; u_e_branch = []
        daughterFlag = True
        countBranch = 0
        for d in decayData:
            if d[0] == "DAU" and d[1] == Daughter_vec[dd]:
                daughterFlag = True
            if d[0] == "DAU" and d[1] != Daughter_vec[dd]:
                daughterFlag = False
            
            if daughterFlag:
                # particle_i = []; p_branch_i=[]; u_p_branch_i=[];  e_branch_i=[];   u_e_branch_i=[];  LevelDaughter_i=[]
                if d[0] =="COM":
                    if "Branch" in d[1] or "Level" in d[1]:
                        print(d[1])
                        if countBranch % 2 == 1:
                            p_branch.append(p_branch_i); u_p_branch.append(u_p_branch_i); # Branching ratio of the decay branch
                            e_branch.append(e_branch_i); u_e_branch.append(u_e_branch_i); # Energy of the decay branch (kinetic energy of the particle)
                            LevelDaughter.append(LevelDaughter_i); particle.append(particle_i)
                            
                            particle_i = []; p_branch_i=[]; u_p_branch_i=[];  e_branch_i=[];   u_e_branch_i=[];  LevelDaughter_i=[]
                        else:
                            particle_i = []; p_branch_i=[]; u_p_branch_i=[];  e_branch_i=[];   u_e_branch_i=[];  LevelDaughter_i=[]
                    
                    
                        countBranch += 1    
                    # p_branch=
                
                
                if d[0] == "ALP" or d[0] == "BEM" or d[0] == "BEP" or d[0] == "CK" or d[0] == "CL" or d[0] == "CL1" or d[0] == "CL2" or d[0] == "CM" or d[0] == "CN":                         # Read information on the decay branch
                    if d[0] == "ALP": particle_i.append("alpha")
                    if d[0] == "BEP": particle_i.append("beta+")
                    if d[0] == "BEM": particle_i.append("beta")
                    if d[0] == "CK": particle_i.append("Atom_K")
                    if d[0] == "CL": particle_i.append("Atom_L")
                    if d[0] == "CL1": particle_i.append("Atom_L1")
                    if d[0] == "CL2": particle_i.append("Atom_L2")
                    if d[0] == "CM": particle_i.append("Atom_M")
                    if d[0] == "CN": particle_i.append("Atom_N")
                    if d[2] == "  ": d[2]=0
            
                    p_branch_i.append(float(d[1])); u_p_branch_i.append(float(d[2])); # Branching ratio of the decay branch
                    e_branch_i.append(float(d[4])); u_e_branch_i.append(float(d[5])); # Energy of the decay branch (kinetic energy of the particle)
                    LevelDaughter_i.append(int(d[3]))                               # Level fed in daughter
                    
                
        
        """
        LOOP IN NLEVEL (Levels for each daughter, starting in NLEVEL, ending in 1)
        """
        levelEnergy = []; fromBranch = []; lineL=[]; listTran = []; levelNumber = []
        transitionType = []; prob = []; u_prob = []; e_trans = []; u_e_trans = []; next_level = []
        transitionType_i = []; prob_i = []; u_prob_i = []; e_trans_i = []; u_e_trans_i = []; next_level_i = []; levelEnergy_i = []
        for d in decayData:
          if d[0] == "DAU" and d[1] == Daughter_vec[dd]:
                daughterFlag = True
          if d[0] == "DAU" and d[1] != Daughter_vec[dd]:
                daughterFlag = False
          if daughterFlag:
              if d[0] == "LED":
                # record transition details of the previous level
                 transitionType.append(transitionType_i); prob.append(prob_i); u_prob.append(u_prob_i); e_trans.append(e_trans_i); u_e_trans.append(u_e_trans_i); next_level.append(next_level_i);levelEnergy.append(levelEnergy_i)
            #    if int(d[6]) in LevelDaughter :  # Read the information on the possible energy levels after the emission of the alpha particle 
                 levelEnergy_i.append(float(d[1]))   #levelEnergy.append(float(d[1]));     # Energie (rounded) of the level
                 listTran.append(int(d[3]))           # Number of transitions that depopulate this level
                 levelNumber.append(int(d[6]))        # Level number
                 transitionType_i = []; next_level_i = []; levelEnergy_i = [];prob_i = []; u_prob_i = []; e_trans_i = []; u_e_trans_i = [];
                  
              """
              LOOP IN NTRANS (Transitions depopulating this level)
              """
              if d[0] == "GA" or d[0] == "EK" or d[0] == "EL1" or d[0] == "EL2" or d[0] == "EL3" or d[0] == "EM" or d[0] == "EN":
                  if d[1] == '  ' or d[1] == '   ': d[1] = 0
                  if d[2] == '  ': d[2] = 0
                  if d[4] == '  ': d[4] = 0
                  transitionType_i.append(d[0])                                          # Read the type of transtion
                  prob_i.append(float(d[1])); u_prob_i.append(float(d[2]))                 # Read the emission probability of the transition
                  e_trans_i.append(float(d[3])); u_e_trans_i.append(float(d[4]))           # Read the energy of the transition
                  next_level_i.append(int(d[5]))                                         # Read the level fed by this transition
            
        transitionType.append(transitionType_i); prob.append(prob_i); u_prob.append(u_prob_i); e_trans.append(e_trans_i); u_e_trans.append(u_e_trans_i); next_level.append(next_level_i);levelEnergy.append(levelEnergy_i)
    
        out.append([particle, p_branch, e_branch, LevelDaughter, levelNumber, prob, levelEnergy, transitionType, e_trans, next_level, Q_value_vec[dd], Daughter_vec[dd], Pdaughter_vec[dd]])
    return out
#out = readPenNuc('Am-242')
#print(len(out),out)
# tic()
# readPenNuc("Co-60")
# toc() # 0.016 s

if absolutePath: file_pennuc = 'G:\Python_modules\Jialin\Code\decayData\\All-nuclides_PenNuc.zip'
else: file_pennuc = "decayData//All-nuclides_PenNuc.zip"

z_PenNuc = zf.ZipFile(file_pennuc)

def readPenNuc2(rad,z1=z_PenNuc):
    '''
     =========
     PARAMETRE
     =========
     rad -- type: str (par exemple: "Am-241") -- radionucléide 

     ======
     RETURN
     ======
     daughter -- indice 0 -- des noyaux fils -- len = nb de noyaux fils
     prob_daug -- indice 1 -- des probabilités de noyaux fils -- len = nb de noyaux fils
     energy_Q -- indice 2 -- des énergies de désintégrations -- len = nb de noyaux fils

     desin_type_tot -- indice 3 -- des types de désintégrations/particules émis
         len = nb de noyaux fils 
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch

     desin_energy_tot -- indice 4 -- des énergies de désintégrations/énergies de patricules émis
         len = nb de noyaux fils 
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch

     desin_prob_tot -- indice 5 -- des probabilités de désintégrations
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des probabilités de désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch
     
     desin_level_tot -- indice 6 -- des niveaux atteints après des désintégrations
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux après des désintégrations de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch
     
     prob_branch_tot -- indice 7 -- probabilités de chaque branch
         len = nb de noyaux fils
         sous-list -- des probabilités de branchs de noyau fil -- len de sous-list = nb de branch de chaque fil
         
     tran_type_tot -- indice 8 -- transitions possibles 
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_energy_tot -- indice 9 -- énergies de transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_prob_tot -- indice 10 -- probabilités de transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des probabilités de transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_level_tot -- indice 11 -- niveaux de branch correspondants
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux de chaque branch avant des transitions -- len de sous-list de sous-list = 1
     
     tran_level_end_tot -- indice 12 -- niveaux après des transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux après des transitions de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     level_energy_tot -- indice 13 -- énergies de niveaux
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de niveaux de chaque branch -- len de sous-list de sous-list = 1

     prob_tran_tot -- indice 14 -- la somme de transition de chaque branch
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de niveaux de chaque branch -- len de sous-list de sous-list = 1
     '''
    doc = rad + ".PenNuc.txt"
    with z1.open(doc) as file_P:
        decayData = file_P.readlines()

    for i in range(np.size(decayData)):
        decayData[i] = str(decayData[i])
        decayData[i] = decayData[i].replace("b'","")
        decayData[i] = decayData[i].replace("\\r\\n","")
        decayData[i] = decayData[i].replace("'","") 

    for il in range(len(decayData)):
        if "NDA " in decayData[il]: decayData[il] = decayData[il].replace("NDA ","NDA; ") 
        if "DAU " in decayData[il]: decayData[il] = decayData[il].replace("DAU ","DAU; ") 
        if "DDE " in decayData[il]: decayData[il] = decayData[il].replace("DDE ","DDE; ")
        if "Q " in decayData[il]: decayData[il] = decayData[il].replace("Q ","Q; ")
        if "ALP " in decayData[il]: decayData[il] = decayData[il].replace("ALP ","ALP; ")
        if "CK " in decayData[il]: decayData[il] = decayData[il].replace("CK ","CK; ")
        if "CL " in decayData[il]: decayData[il] = decayData[il].replace("CL ","CL; ")
        if "CO " in decayData[il]: decayData[il] = decayData[il].replace("CO ","CO; ")
        if "CL1 " in decayData[il]: decayData[il] = decayData[il].replace("CL1 ","CL1; ")
        if "CL2 " in decayData[il]: decayData[il] = decayData[il].replace("CL2 ","CL2; ")
        if "CL3 " in decayData[il]: decayData[il] = decayData[il].replace("CL3 ","CL3; ")
        if "CN " in decayData[il]: decayData[il] = decayData[il].replace("CN ","CN; ")
        if "CM " in decayData[il]: decayData[il] = decayData[il].replace("CM ","CM; ")
        if "BEM " in decayData[il]: decayData[il] = decayData[il].replace("BEM ","BEM; ")
        if "BEP " in decayData[il]: decayData[il] = decayData[il].replace("BEP ","BEP; ")
        if "LED " in decayData[il]: decayData[il] = decayData[il].replace("LED ","LED; ")
        if "GA " in decayData[il]: decayData[il] = decayData[il].replace("GA ","GA; ")
        if "EK " in decayData[il]: decayData[il] = decayData[il].replace("EK ","EK; ")
        if "EL " in decayData[il]: decayData[il] = decayData[il].replace("EL ","EL; ")
        if "EL1 " in decayData[il]: decayData[il] = decayData[il].replace("EL1 ","EL1; ")
        if "EL2 " in decayData[il]: decayData[il] = decayData[il].replace("EL2 ","EL2; ")
        if "EL3 " in decayData[il]: decayData[il] = decayData[il].replace("EL3 ","EL3; ")
        if "EM " in decayData[il]: decayData[il] = decayData[il].replace("EM ","EM; ")
        if "EN " in decayData[il]: decayData[il] = decayData[il].replace("EN ","EN; ")
        if "COM " in decayData[il]: decayData[il] = decayData[il].replace("COM ","COM; ")
        decayData[il] = decayData[il].split(';')

    for a1 in decayData:
        for a2 in range(len(a1)):
            a1[a2] = a1[a2].strip()

    '''
     ========================
     Repérer chaque noyau fil
     ========================

     daughter -- noyau(x) fil(s)
     posi_daug -- l'indice de démarcation de noyau fil
     posi_branch -- l'indice de démarcation de chaque branch
     posi_tran -- l'indice de démarcation de transition
     prob_daug -- probabilité de produire des noyaux fils
     nb_branch -- nombre de branch possible au dessus de l'état fonda (n>0)
     energy_Q -- l'énergie de désintégration

    '''    
    daughter = [];posi_daug = [];prob_daug=[];nb_branch=[];energy_Q=[];
    end = len(decayData)
    for indice,line in enumerate(decayData):
        if "NDA" == line[0]:
            nb_daug = int(line[1])
        if "DAU" == line[0]:
            daughter.append(line[1].replace(" ",""))
        if "COM" == line[0] and "Daughter" in line[1]:
            posi_daug.append(indice)
        if "Q" == line[0]:
            energy_Q.append(float(line[1]))
        if "DDE" == line[0]:
            prob_daug.append(float(line[1]))
            nb_branch.append(int(line[-2]))
    '''
     ==========
     LOOP START
     ==========

    ''' 
    posi_end=[]
    desin_type_tot=[];desin_energy_tot=[];desin_prob_tot=[];desin_level_tot=[]
    tran_type_tot=[];tran_energy_tot=[];tran_prob_tot=[];tran_level_end_tot=[]; 
    tran_level_tot=[];level_energy_tot=[]
    prob_branch_tot=[];prob_tran_tot=[] 

    '''
     =============
     LOOP DAUGHTER 
     =============

    '''
    for i1 in range(nb_daug):
        start_p = posi_daug[i1]
        if i1+1 == nb_daug:
            end_p = end
        else:
            end_p = posi_daug[i1+1]

        posi_end_i = []
        for i2 in range(start_p,end_p):
            if "COM" == decayData[i2][0] and "Branch" in decayData[i2][1]:
                posi_end_i.append(i2)
            if "COM" == decayData[i2][0] and "Level" in decayData[i2][1]:
                posi_end_i.append(i2)
        if end_p == end:
            posi_end_i.append(end)
        else:
            posi_end_i.append(posi_daug[i1+1])
        posi_end.append(posi_end_i)

        '''
         ==========================
         LOOP Branch and Transition
         ==========================
        '''
        desin_type_daug=[];desin_energy_daug=[];desin_prob_daug=[];desin_level_daug=[];
        tran_type_daug=[];tran_energy_daug=[];tran_prob_daug=[];tran_level_end_daug=[];
        tran_level_daug=[];level_energy_daug=[]
        prob_branch_daug=[];prob_tran_daug=[]

        for i3 in range(len(posi_end_i)-1):
            start_p1 = posi_end_i[i3]
            end_p1 = posi_end_i[i3+1]
            branch = False
            transition = False
            if "COM" == decayData[start_p1][0] and "Branch" in decayData[start_p1][1]:
                branch=True
            if "COM" == decayData[start_p1][0] and "Level" in decayData[start_p1][1]:
                transition = True

            '''
             ====================================
             LOOP EACH BLOCK OF BRANCH/TRANSITION
             ====================================
            ''' 
            tran_type_b=[];tran_prob_b=[];tran_energy_b=[]; tran_level_end_b=[];
            tran_level_b=[];level_energy_b=[]
            desin_type_b=[];desin_energy_b=[];desin_prob_b=[];desin_level_b=[];
             
            for i4 in decayData[start_p1+1:end_p1]:
                if start_p1+1 == end_p1:
                    break
                if branch:
                    if "ALP" == i4[0]:
                        desin_type_b.append("alpha")
                    if "BEP" == i4[0]:
                        desin_type_b.append("beta+")
                    if "BEM" == i4[0]:
                        desin_type_b.append("beta")
                    if "CK" == i4[0]:
                        desin_type_b.append("Atom_K")
                    if "CL" == i4[0]:
                        desin_type_b.append("Atom_L")
                    if "CL1" == i4[0]:
                        desin_type_b.append("Atom_L1")
                    if "CL2" == i4[0]:
                        desin_type_b.append("Atom_L2")
                    if "CL3" == i4[0]:
                        desin_type_b.append("Atom_L3")
                    if "CM" == i4[0]:
                        desin_type_b.append("Atom_M")
                    if "CN" == i4[0]:
                        desin_type_b.append("Atom_N")
                    if "CO" == i4[0]:
                        desin_type_b.append("Atom_O")
                    desin_prob_b.append(float(i4[1]))
                    desin_level_b.append(int(i4[3]))
                    desin_energy_b.append(float(i4[4]))
                if transition:
                    if i4[1] == '' or i4[1] == ' ': i4[1] = 0
                    if len(i4)>2 and i4[2] == '': i4[2] = 0
                    if len(i4)>4 and i4[4] == '': i4[4] = 0
                    if len(i4)>5 and i4[5] == '': i4[5] = 0
                    if "LED" == i4[0]:
                        tran_level_b.append(int(i4[-1]))
                        level_energy_b.append(float(i4[1]))
                    if i4[0] == "GA" or i4[0] == "EK" or i4[0] == "EL" or i4[0] == "EL1" or i4[0] == "EL2" or i4[0] == "EL3" or i4[0] == "EM" or i4[0] == "EN":
                        #print(i4)
                        tran_type_b.append(i4[0])
                        tran_prob_b.append(float(i4[1]))
                        tran_energy_b.append(float(i4[3]))
                        tran_level_end_b.append(int(i4[5]))
            if branch:
                desin_type_daug.append(desin_type_b)
                desin_energy_daug.append(desin_energy_b)
                desin_prob_daug.append(desin_prob_b)
                desin_level_daug.append(desin_level_b)
             
            if transition:
                tran_type_daug.append(tran_type_b)
                tran_energy_daug.append(tran_energy_b)
                tran_prob_daug.append(tran_prob_b)
                tran_level_end_daug.append(tran_level_end_b)
                tran_level_daug.append(tran_level_b)
                level_energy_daug.append(level_energy_b)

            if len(desin_prob_b)>0:
                desin_prob_array = np.array(desin_prob_b)
                prob_branch_i = np.sum(desin_prob_array)
                if prob_branch_i >= 1:
                    prob_branch_i = 1
                prob_branch_daug.append(prob_branch_i)
            elif branch and len(desin_prob_b)==0:
                prob_branch_daug.append(0)

            if len(tran_prob_b)>0:
                tran_prob_array = np.array(tran_prob_b)
                prob_tran_i = np.sum(tran_prob_array)
                if prob_tran_i >= 1:
                    prob_tran_i = 1
                prob_tran_daug.append(prob_tran_i)
            elif transition and len(tran_prob_b)==0:
                prob_tran_daug.append(0)

        tran_type_daug.append([])
        tran_prob_daug.append([])
        tran_energy_daug.append([])
        tran_level_end_daug.append([])
        tran_level_daug.append([])
        level_energy_daug.append([])
        prob_tran_daug.append(0)

        desin_type_tot.append(desin_type_daug)
        desin_energy_tot.append(desin_energy_daug)
        desin_prob_tot.append(desin_prob_daug)
        desin_level_tot.append(desin_level_daug)

        tran_type_tot.append(tran_type_daug)
        tran_energy_tot.append(tran_energy_daug)
        tran_prob_tot.append(tran_prob_daug)
        tran_level_end_tot.append(tran_level_end_daug)
        tran_level_tot.append(tran_level_daug)
        level_energy_tot.append(level_energy_daug)
        prob_branch_tot.append(prob_branch_daug)
        prob_tran_tot.append(prob_tran_daug)
    out = [daughter,prob_daug,energy_Q,desin_type_tot,desin_energy_tot,desin_prob_tot,desin_level_tot,prob_branch_tot,tran_type_tot,tran_energy_tot,tran_prob_tot,tran_level_tot,tran_level_end_tot,level_energy_tot,prob_tran_tot]
    return out
#tic()
#o = readPenNuc2("Sb-127")
#toc()
#print(o[-4])


def readPenNuc1(rad):
     '''
     =========
     PARAMETRE
     =========
     rad -- type: str (par exemple: "Am-241") -- radionucléide 

     ======
     RETURN
     ======
     daughter -- indice 0 -- des noyaux fils -- len = nb de noyaux fils
     prob_daug -- indice 1 -- des probabilités de noyaux fils -- len = nb de noyaux fils
     energy_Q -- indice 2 -- des énergies de désintégrations -- len = nb de noyaux fils

     desin_type_tot -- indice 3 -- des types de désintégrations/particules émis
         len = nb de noyaux fils 
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch

     desin_energy_tot -- indice 4 -- des énergies de désintégrations/énergies de patricules émis
         len = nb de noyaux fils 
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch

     desin_prob_tot -- indice 5 -- des probabilités de désintégrations
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des probabilités de désintégrations possibles de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch
     
     desin_level_tot -- indice 6 -- des niveaux atteints après des désintégrations
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux après des désintégrations de chaque branch -- len de sous-list de sous-list = nb de type de désintégrations de chaque branch
     
     prob_branch_tot -- indice 7 -- probabilités de chaque branch
         len = nb de noyaux fils
         sous-list -- des probabilités de branchs de noyau fil -- len de sous-list = nb de branch de chaque fil
         
     tran_type_tot -- indice 8 -- transitions possibles 
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_energy_tot -- indice 9 -- énergies de transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_prob_tot -- indice 10 -- probabilités de transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des probabilités de transitions possibles de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     tran_level_tot -- indice 11 -- niveaux de branch correspondants
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux de chaque branch avant des transitions -- len de sous-list de sous-list = 1
     
     tran_level_end_tot -- indice 12 -- niveaux après des transitions
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des niveaux après des transitions de chaque branch -- len de sous-list de sous-list = nb de type de transitions de chaque branch
     
     level_energy_tot -- indice 13 -- énergies de niveaux
         len = nb de noyaux fils
         sous-list -- des branchs possibles de noyau fil -- len de sous-list = nb de branch de chaque fil
         sous-list de sous-list -- des énergies de niveaux de chaque branch -- len de sous-list de sous-list = 1
     '''

     url = "http://www.lnhb.fr/nuclides/"+rad+".PenNuc.txt"
     file = rq.urlopen(url)
     
     # Format the data 
     decayData = []
     for line in file:
        decayData.append(line.decode("utf-8"))
        if "NDA " in decayData[-1]:    decayData[-1] = decayData[-1].replace("NDA ", "NDA; ") # number of daughter
        if "DAU " in decayData[-1]:    decayData[-1] = decayData[-1].replace("DAU ", "DAU; ") # daughter
        if "DDE " in decayData[-1]:    decayData[-1] = decayData[-1].replace("DDE ", "DDE; ") # daughter description : probability of disintegration to NDA, uncertainty, number of excited levels of NDA; number of branche to NDA
        if "Q " in decayData[-1]:      decayData[-1] = decayData[-1].replace("Q ", "Q; ")     # total energy of the branch, uncertainty 
        if "ALP " in decayData[-1]:    decayData[-1] = decayData[-1].replace("ALP ", "ALP; ") # type of the disintegration = alpha
        if "CK " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CK ", "CK; ") # type of the disintegration = Electron Capture K shell
        if "CL " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL ", "CL; ") # type of the disintegration = Electron Capture L shell
        if "CL1 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL1 ", "CL1; ") # type of the disintegration = Electron Capture L shell 1
        if "CL2 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL2 ", "CL2; ") # type of the disintegration = Electron Capture L shell 2
        if "CL3 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CL3 ", "CL3; ") # type of the disintegration = Electron Capture L shell 3
        if "CM " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CM ", "CM; ") # type of the disintegration = Electron Capture M shell
        if "CN " in decayData[-1]:    decayData[-1] = decayData[-1].replace("CN ", "CN; ") # type of the disintegration = Electron Capture N shell
        if "BEM " in decayData[-1]:    decayData[-1] = decayData[-1].replace("BEM ", "BEM; ") # type disintegration-beta- :branching ratio, uncertainty, level fed in daughter, energy, uncertainty,prohibition factor for beta emission
        if "BEP " in decayData[-1]:    decayData[-1] = decayData[-1].replace("BEP ", "BEP; ") # type disintegration-beta+ :branching ratio, uncertainty, level fed in daughter, energy, uncertainty,prohibition factor for beta emission
        if "LED " in decayData[-1]:    decayData[-1] = decayData[-1].replace("LED ", "LED; ") # level description:energy of level,uncertainty, number of transitions that depopulate this level,level time,uncertainty,level number
        if "GA " in decayData[-1]:     decayData[-1] = decayData[-1].replace("GA ", "GA; ")   # type transition: gamma
        if "EK " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EK ", "EK; ")   # type transition: internal conversion -- electron K
        if "EL " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL ", "EL; ") # type transition: internal conversion -- electron L
        if "EL1 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL1 ", "EL1; ") # type transition: internal conversion -- electron L1
        if "EL2 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL2 ", "EL2; ") # type transition: internal conversion -- electron L2
        if "EL3 " in decayData[-1]:    decayData[-1] = decayData[-1].replace("EL3 ", "EL3; ") # type transition: internal conversion -- electron L3
        if "EM " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EM ", "EM; ")   # type transition: internal conversion -- electron M
        if "EN " in decayData[-1]:     decayData[-1] = decayData[-1].replace("EN ", "EN; ")   # type transition: internal conversion -- electron N
        if "COM " in decayData[-1]:    decayData[-1] = decayData[-1].replace("COM ", "COM; ") 
        decayData[-1]=decayData[-1].split(";")
        if "\r\n" in decayData[-1][-1]:  decayData[-1][-1] = decayData[-1][-1].replace("\r\n", " ")

     '''
     ========================
     Repérer chaque noyau fil
     ========================

     daughter -- noyau(x) fil(s)
     posi_daug -- l'indice de démarcation de noyau fil
     posi_branch -- l'indice de démarcation de chaque branch
     posi_tran -- l'indice de démarcation de transition
     prob_daug -- probabilité de produire des noyaux fils
     nb_branch -- nombre de branch possible au dessus de l'état fonda (n>0)
     energy_Q -- l'énergie de désintégration

     '''    
     daughter = [];posi_daug = [];prob_daug=[];nb_branch=[];energy_Q=[];
     end = len(decayData)
     for indice,line in enumerate(decayData):
         if "NDA" == line[0]:
             nb_daug = int(line[1])
         if "DAU" == line[0]:
             daughter.append(line[1].replace(" ",""))
         if "COM" == line[0] and "Daughter" in line[1]:
             posi_daug.append(indice)
         if "Q" == line[0]:
             energy_Q.append(float(line[1]))
         if "DDE" == line[0]:
             prob_daug.append(float(line[1]))
             nb_branch.append(int(line[-2]))

     '''
     ==========
     LOOP START
     ==========

     '''  
     posi_end=[]
     desin_type_tot=[];desin_energy_tot=[];desin_prob_tot=[];desin_level_tot=[]
     tran_type_tot=[];tran_energy_tot=[];tran_prob_tot=[];tran_level_end_tot=[]; 
     tran_level_tot=[];level_energy_tot=[]
     prob_branch_tot=[]

     '''
     =============
     LOOP DAUGHTER 
     =============
     '''
     for i1 in range(nb_daug):
        start_p = posi_daug[i1]
        if i1+1 == nb_daug:
            end_p = end
        else:
            end_p = posi_daug[i1+1]

        posi_end_i = []
        for i2 in range(start_p,end_p):
            if "COM" == decayData[i2][0] and "Branch" in decayData[i2][1]:
                posi_end_i.append(i2)
            if "COM" == decayData[i2][0] and "Level" in decayData[i2][1]:
                posi_end_i.append(i2)
        if end_p == end:
            posi_end_i.append(end)
        else:
            posi_end_i.append(posi_daug[i1+1])
        posi_end.append(posi_end_i)
     
        '''
        ==========================
        LOOP Branch and Transition
        ==========================
        '''

        desin_type_daug=[];desin_energy_daug=[];desin_prob_daug=[];desin_level_daug=[];
        tran_type_daug=[];tran_energy_daug=[];tran_prob_daug=[];tran_level_end_daug=[];
        tran_level_daug=[];level_energy_daug=[]
        prob_branch_daug=[]

        for i3 in range(len(posi_end_i)-1):
            start_p1 = posi_end_i[i3]
            end_p1 = posi_end_i[i3+1]
            branch = False
            transition = False
            if "COM" == decayData[start_p1][0] and "Branch" in decayData[start_p1][1]:
                branch=True
            if "COM" == decayData[start_p1][0] and "Level" in decayData[start_p1][1]:
                transition = True

            '''
            ====================================
            LOOP EACH BLOCK OF BRANCH/TRANSITION
            ====================================
            ''' 
            tran_type_b=[];tran_prob_b=[];tran_energy_b=[]; tran_level_end_b=[];
            tran_level_b=[];level_energy_b=[]
            desin_type_b=[];desin_energy_b=[];desin_prob_b=[];desin_level_b=[];
             
            for i4 in decayData[start_p1+1:end_p1]:
                
                if start_p1+1 == end_p1:
                    break
                if branch:
                    if "ALP" == i4[0]:
                        desin_type_b.append("alpha")
                    if "BEP" == i4[0]:
                        desin_type_b.append("beta+")
                    if "BEM" == i4[0]:
                        desin_type_b.append("beta")
                    if "CK" == i4[0]:
                        desin_type_b.append("Atom_K")
                    if "CL" == i4[0]:
                        desin_type_b.append("Atom_L")
                    if "CL1" == i4[0]:
                        desin_type_b.append("Atom_L1")
                    if "CL2" == i4[0]:
                        desin_type_b.append("Atom_L2")
                    if "CL3" == i4[0]:
                        desin_type_b.append("Atom_L3")
                    if "CM" == i4[0]:
                        desin_type_b.append("Atom_M")
                    if "CN" == i4[0]:
                        desin_type_b.append("Atom_N")
                    desin_prob_b.append(float(i4[1]))
                    desin_level_b.append(int(i4[3]))
                    desin_energy_b.append(float(i4[4]))

                if transition:
                    if i4[1] == '  ' or i4[1] == '   ': i4[1] = 0
                    if len(i4)>2 and i4[2] == '  ': i4[2] = 0
                    if len(i4)>4 and i4[4] == '  ': i4[4] = 0
                    if len(i4)>5 and i4[5] == '  ': i4[5] = 0
                    if "LED" == i4[0]:
                        tran_level_b.append(int(i4[-1]))
                        level_energy_b.append(float(i4[1]))
                    if i4[0] == "GA" or i4[0] == "EK" or i4[0] == "EL" or i4[0] == "EL1" or i4[0] == "EL2" or i4[0] == "EL3" or i4[0] == "EM" or i4[0] == "EN":
                        tran_type_b.append(i4[0])
                        tran_prob_b.append(float(i4[1]))
                        tran_energy_b.append(float(i4[3]))
                        tran_level_end_b.append(float(i4[5]))

            if branch:
                desin_type_daug.append(desin_type_b)
                desin_energy_daug.append(desin_energy_b)
                desin_prob_daug.append(desin_prob_b)
                desin_level_daug.append(desin_level_b)
             
            if transition:
                tran_type_daug.append(tran_type_b)
                tran_energy_daug.append(tran_energy_b)
                tran_prob_daug.append(tran_prob_b)
                tran_level_end_daug.append(tran_level_end_b)
                tran_level_daug.append(tran_level_b)
                level_energy_daug.append(level_energy_b)

            if len(desin_prob_b)>0:
                desin_prob_array = np.array(desin_prob_b)
                prob_branch_i = np.sum(desin_prob_array)
                if prob_branch_i >= 1:
                    prob_branch_i = 1
                prob_branch_daug.append(prob_branch_i)
            elif branch and len(desin_prob_b)==0:
                prob_branch_daug.append(0)

        tran_type_daug.append([])
        tran_prob_daug.append([])
        tran_energy_daug.append([])
        tran_level_end_daug.append([])
        tran_level_daug.append([])
        level_energy_daug.append([])

        desin_type_tot.append(desin_type_daug)
        desin_energy_tot.append(desin_energy_daug)
        desin_prob_tot.append(desin_prob_daug)
        desin_level_tot.append(desin_level_daug)

        tran_type_tot.append(tran_type_daug)
        tran_energy_tot.append(tran_energy_daug)
        tran_prob_tot.append(tran_prob_daug)
        tran_level_end_tot.append(tran_level_end_daug)
        tran_level_tot.append(tran_level_daug)
        level_energy_tot.append(level_energy_daug)
        prob_branch_tot.append(prob_branch_daug)

     out = [daughter,prob_daug,energy_Q,desin_type_tot,desin_energy_tot,desin_prob_tot,desin_level_tot,prob_branch_tot,tran_type_tot,tran_energy_tot,tran_prob_tot,tran_level_tot,tran_level_end_tot,level_energy_tot]
     return out

#tic()
#o = readPenNuc1("Ag-110m")
#print(o[-4])
#toc()
#===============================================================================================================
'''
rad = "Am-244m"
out1 = readPenNuc1(rad)
print(rad)
print("    ")
print("DAU",out1[0])
print("    ")
print("prob_daug",out1[1])
print("    ")
print("Q",out1[2])
print("    ")
print("desin_type",out1[3],len(out1[3]))
print("    ")
print("desin-prob",out1[5],len(out1[5]))
print("    ")
print("desin-level",out1[6],len(out1[6]))
print("    ")
print("prob-branch",out1[7])
print("    ")
print("len-tran-type",len(out1[8]))
for i in range(len(out1[8])):
    print("fil",i)
    for i1,p1 in enumerate(out1[8][i]):
        print(p1,len(p1))
    print("energy")
    for i2,p2 in enumerate(out1[9][i]):
        print(p2,len(p2))
    print("tran-prob")
    for i4,p4 in enumerate(out1[10][i]):
        print(p4)
    print("end_level")
    for i3,p3 in enumerate(out1[12][i]):
        print(p3,len(p3))
'''
#print(readPenNuc1('Tc-99m'))
#out = readPenNuc1('Tc-99m')
#print(sampling(out[1]))
#================================== StoppingPower for alpha particle ===========================================

if absolutePath: f_alpha = open('G:\Python_modules\Jialin\Code\Quenching\\alpha_toulene.txt')
else: f_alpha = open('Quenching/alpha_toulene.txt')
data_ASTAR = f_alpha.readlines()
f_alpha.close()
energy_alph = []
dEdx_alph = []
for i in range(np.size(data_ASTAR)):
    data_ASTAR[i] = data_ASTAR[i].split()
    for j in range(2):
        data_ASTAR[i][j] = float(data_ASTAR[i][j])*1e3  # dEdx from MeV.cm2/g to keV.cm2/g; energy from MeV to keV
    energy_alph.append(data_ASTAR[i][0])
    dEdx_alph.append(data_ASTAR[i][1])

def stoppingpowerA(e,rho=0.96,energy_alpha=energy_alph,dEdx_alpha=dEdx_alph):
    """
    Estimation of the stopping power of alpha particles using tabulated values form the ASTAR code
    ref: https://dx.doi.org/10.18434/T4NC7P
    
    Parameters
    ----------
    e : float
        energy of the alpha particle in keV.
    rho : float, optional
        density of the source in g.cm-3. The default is 0.96.
    energy_alpha : list, optional
        the list of energy (in keV) for which the stopping power was calculated with ASTAR. The default is energy_alph.
    dEdx_alpha : list, optional
        the list of stopping powers (in keV.cm2/g) associated with the energy vector. The default is dEdx_alph.

    Returns
    -------
    float
        Interpolated ASTAR estimation of the stopping power.

    """

    energy_alpha = np.array(energy_alpha)
    dEdx_alpha = np.array(dEdx_alpha)
    dEdx = np.interp(e,energy_alpha ,dEdx_alpha)   
    return dEdx*rho                        #unit keV.cm-1

# tic()
# stoppingpowerA(5000,rho=0.96,energy_alpha=energy_alph,dEdx_alpha=dEdx_alph)
# toc() # 0 s

#===============================================================================================

#========================   Nouveau modèle pour calculer le pouvoir d'arrête d'électron ========


if absolutePath: file_TanXia=open('G:\Python_modules\Jialin\Code\Quenching\\TandataUG.txt', "r")
else: file_TanXia=open('Quenching/TandataUG.txt', "r")
data_TanXia=file_TanXia.read(); file_TanXia.close()
data_TanXia=data_TanXia.split("\n"); data_TanXia_f = np.empty(len(data_TanXia))
for i, x in enumerate(data_TanXia):
  if i<len(data_TanXia)-1: data_TanXia_f[i]=float(x)

def stoppingpower(e,rho=0.96,Z=5.2,A=11.04,emin=0,file=data_TanXia_f):
    """
    The stopping power of electrons between 20 keV and 1000 keV is a mixture of a radiative loss model [1], and a collision model [2] that has been validated agaisnt the NIST model ESTAR [3] recommanded by the ICRU Report 37 [4].
    At low energy - between 10 eV and 20 keV - the model from Tan and Xia [5] is implemented.
    Refs:
        [1] https://doi.org/10.1016/0020-708x(82)90244-7
        [2] https://www.ijstr.org/final-print/jan2017/Calculations-Of-Stopping-Power-And-Range-Of-Electrons-Interaction-With-Different-Material-And-Human-Body-Parts.pdf
        [3] https://dx.doi.org/10.18434/T4NC7P
        [4] ICRU Report 37, Stopping Powers for Electrons and Positrons
        [5] https://doi.org/10.1016/j.apradiso.2011.08.012
        
    Parameters
    ----------
    e : float
        Energy of the electron in eV.
    rho : float, optional
        density of the source in g.cm-3. The default is 0.96.
    Z : float, optional
        mean charge number of the source. The default is 5.2.
    A : float, optional
        mean mass number of the source. The default is 11.04.
    emin : float, optional
        the minimal energy to consider. The default is 0.
    file : list, optional
        tabulated data form the Tan and Xia model. The default is data_TanXia_f.

    Returns
    -------
    dEdx : float
        Calculated stopping power in MeV.cm-1.

    """
    # e:eV ;rho: g.cm-3
    mc_2 = 0.511 #MeV
    I = 65e-6 #MeV
    NA = 6.02e23
    ahc = 1.437e-13   #MeV.cm
    if e>=20000:
        e1 = e*1e-6 #MeV
        gamma = (e1+mc_2)/mc_2
        gamma_2 = gamma*gamma
        beta = np.sqrt(1-(1/gamma_2))
        beta_2 = beta**2
        tau = e1/mc_2
        terma = np.log(tau**2*(tau+2)/2)
        termb = 1+tau*tau/8-(2*tau+1)*np.log(2)
        termc = (tau+1)**2
        B0 = terma + termb/termc
        sc = 0.1535/beta_2*Z/A*(B0-2*np.log(I/mc_2))  
        term3 = NA*(Z**2)*rho*(e1+mc_2)/(137*(mc_2**2)*A)
        term4 = 4* np.log(2*gamma) -4/3
        sr = (ahc**2)*term3*term4
        #if T<1:sr=0
        dEdx = (sc + sr)*rho  #MeV.cm-1
    else:
            if e>emin:
                dEdx=float(file[int(e)]) #MeV.cm-1
            else:
                dEdx=0
    if dEdx<0:
        dEdx=0
    return dEdx    

# tic()
# stoppingpower(100,rho=0.96,Z=5.2,A=11.04,emin=0,file=data_TanXia_f)
# toc() # 0 s
#===================================================================================================

#===================================================================================================
#============================tracer stopping power of electron and alpha ===========================

'''
e1 = np.linspace(10,1.99e4,10000)
e2 = np.linspace(1.99e4,1e8,20000)
sp1 = []
for i in e1:
    sp1.append(stoppingpower(i))
sp2=[]
for i in e2:
    sp2.append(stoppingpower(i))

ea = np.linspace(1,8e3,4000)
spa = []
for i in ea:
    spa.append(stoppingpowerA(i)*1e-3)
x = ea*1e3
plt.plot(e1,sp1,color='blue',label="pouvoir d'arrête électron")
plt.plot(e2,sp2,color='blue')
plt.plot(x,spa,label="pouvoir d'arrête alpha")
plt.xscale('log')
plt.yscale('log')
plt.title("pouvoir d'arrête d'electron et alpha")
plt.ylabel("pouvoir d'arrête/MeV.cm-1")
plt.xlabel('énergie cinétique/eV')
plt.legend(fontsize=10,loc='best')
plt.savefig('Quenching/stoppingpowerE_A.png')

'''

#=============================================================================================

#====================  Fonction pour lire BetaShape   ========================================
file_betashape = "decayData//All-nuclides_BetaShape.zip"
z_betashape = zf.ZipFile(file_betashape)

def readBetaShape(rad,mode,level,z=z_betashape):
    """
    This funcion reads the beta spectra calculated by the code BetaShape and published in the DDEP web page.
    refs:
        https://doi.org/10.1103/PhysRevC.92.059902
        http://www.lnhb.fr/ddep_wg/

    Parameters
    ----------
    rad : string
        identifier of the radionuclide. e.g. 'Na-22'
    mode : string
        identifier of the decay mode. 'beta-' or 'beta+'
    level : int
        level of the daughter after decay.  0,1,2,3 ....
    Returns
    -------
    e : list
        the energy vector in keV.
    dNdx : list
        the probability density in keV-1.

    """

    Rad = rad.replace('-','')
    name_doc = Rad+'/'+mode+'_'+Rad+'_'+ "trans" + str(level) +'.bs'
    with z.open(name_doc) as file_trans:
        data = file_trans.readlines()

    for i in range(np.size(data)):
        data[i] = str(data[i])
        data[i] = data[i].replace("b'",'')
        data[i] = data[i].replace("\\r\\n",'')
        data[i] = data[i].replace("'",'')
    for i in range(np.size(data)):
        data[i] = data[i].split()
    e = []
    dNdx = []
    
    while [] in data:
        data.remove([])
    
    for i in range(len(data)):
        ind = i
        if data[i][0] == 'E(keV)':break
    
    for j in range(i+1,len(data)):
        e.append(float(data[j][0])) # convert to float
        dNdx.append(float(data[j][1])) # convert to float
    dNdx /= sum(np.asarray(dNdx)) # normalization
    dNdx = list(dNdx)
    return e, dNdx

#tic()
#e,p = readBetaShape('C-11','beta+',0)
#toc()
#print(p,type(p))

# tic()
# e,p = readBetaShape('C-11','beta+',0)
# toc() # 0.016 s > 0s

def readTDCR17spectra(rad):
    file = open("decayData/spectra/spectrumTDCR17_"+rad+".dat")
    data = file.read()
    file.close()
    data = data.split("\n")
    e = []; r = []
    for i in data:
        if "keV" not in i:
            a = i.split("  ")
            if len(a)>1:
                e.append(float(a[1])*1e-3)
                r.append(float(a[2]))
    return e, r


# tic()
# e,r =  readBetaShape("Nb-95m", "beta-", 1)
# # print(e[-1])
# toc() # 0.016 s > 0 s

#=====================================================================================

## Display beta spectra
# rplot = "S-35"
# out = readBetaShape(rplot, "beta-", "tot"); print(sum(out[1]))
# out_t = readTDCR17spectra(rplot); print(sum(out_t[1])) # TDCR17 spectra
# plt.figure("beta spectrum")
# plt.clf()
# plt.plot(out[0],out[1]/(out[0][1]-out[0][0]),label='Beta spectrum - BetaShape',ls='-',color='red',lw=3)
# plt.plot(out_t[0],np.asarray(out_t[1])/(out_t[0][1]-out_t[0][0]),label='Beta spectrum - TDCR17',ls='-',color='blue',lw=3)
# plt.xscale('linear')
# plt.legend(fontsize=12,loc='best')
# plt.xlabel(r'$E$ /keV', fontsize=12)
# plt.ylabel(r'd$N$/d$E$ /keV$^{-1}$',fontsize=12)
# plt.savefig("decayData/spectra/BetaSpectrum_"+rplot+".png")

#=======================================================================================

#============================  Fonction quenching  =====================================

def E_quench_e(e,kB,nE):
    """
    This function calculate the quenched energy of electrons according to the Birks model of scintillation quenching

    Parameters
    ----------
    e : float
        energy of the electron in eV.
    kB : float
        Birks constant in cm/MeV.

    Returns
    -------
    float
        Quenched energy in eV.

    """
    
    e_dis = np.linspace(0,e,nE)
    delta = e_dis[2] - e_dis[1]
    q = 0
    for i in e_dis:
        q += delta/(1+kB*stoppingpower(i))
    return q

def E_quench_a(e,kB,nE): 
    """
    This function calculate the quenched energy alpha particles according to  the Birks model of scintillation quenching

    Parameters
    ----------
    e : float
        energy of the alpha particle in keV.
    kB : float
        Birks constant in cm/keV.

    Returns
    -------
    float
        Quenched energy in keV.

    """

    e_dis = np.linspace(1,e,nE)
    delta = e_dis[2] - e_dis[1]
    q = 0
    for i in e_dis:
        q += delta/(1+kB*stoppingpowerA(i))
    return q

#=========================================================================================



#print(E_quench_e(3.996e5,1e-2,1000)*1e-3)

#======================================================================================

#========================= Tracer les courbes avec kB différents ======================

'''

# s1 = []
# s2 = []
# s3 = []
# x = np.linspace(5,8000,4000) 

# for i in x:
#     s1.append(E_quench_a(i,kB=7e-6)/i)
#     s2.append(E_quench_a(i,kB=1e-5)/i)
#     s3.append(E_quench_a(i,kB=1.4e-5)/i)

# plt.plot(x,s1,label='kB=0.007cm/MeV',color='green',lw=2)
# plt.plot(x,s2,label='kB=0.01cm/MeV',ls=':',color='red',lw=3)
# plt.plot(x,s3,label='kB=0.014cm/MeV')
# plt.xscale('log')
# #plt.yscale('log')
# plt.legend(fontsize=12,loc='best')
# plt.xlabel('E de particule/keV')
# plt.ylabel("énergie d'extinction/E")
# plt.savefig("Quenching/beta 100-10K E_Q sur E.png")

#'''
#============================================================================================

#============================================================================================

#========================= énergie gamma ===================================================
#'''
if absolutePath: 
    fp1 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_1_200k.txt'       #gamma-10ml-1-200keV-niveau 0
    fp2 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_200_2000k.txt'    #gamma-10ml-200-2000keV-niveau 0
    fp3 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_2000_10000k.txt'  #gamma-10ml-2000-10000keV-niveau 0
    fp4 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_16ml-photon_1_200k.txt'       #gamma-10ml-1-200keV-niveau 0
    fp5 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_16ml-photon_200_2000k.txt'       #gamma-10ml-1-200keV-niveau 0
    fe = "G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/E_depose.txt"  
else:
    fp1 = 'MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_1_200k.txt'      #gamma-10ml-1-200keV-niveau 0
    fp2 = 'MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_200_2000k.txt'   #gamma-10ml-200-2000keV-niveau 0
    fp3 = 'MCNP-MATRIX/matrice/fichier/matrice_10ml-photon_2000_10000k.txt' #gamma-10ml-2000-10000keV-niveau 0
    fp4 = 'MCNP-MATRIX/matrice/fichier/matrice_16ml-photon_1_200k.txt'      #gamma-10ml-1-200keV-niveau 0
    fp5 = 'MCNP-MATRIX/matrice/fichier/matrice_16ml-photon_200_2000k.txt'      #gamma-10ml-1-200keV-niveau 0
    fe = "MCNP-MATRIX/matrice/fichier/E_depose.txt"
'''
data1 = f1.readlines()
data2 = f2.readlines()
data3 = f3.readlines()
data_e = fe.readlines()

Matrice1 = np.zeros((1003,200))
Matrice2 = np.zeros((1003,901))
Matrice3 = np.zeros((1003,801))
Matrice_e = np.zeros((1002,3))

for i in range(1002):
    data_e[i] = data_e[i].split()
    for j in range(3):
        Matrice_e[i][j] = float(data_e[i][j])
for i in range(1003):
    data1[i] = data1[i].split()
    data2[i] = data2[i].split()
    data3[i] = data3[i].split()
    for j in range(200):
        Matrice1[i][j] = float(data1[i][j])
    for k in range(901):
        Matrice2[i][k] = float(data2[i][k])
    for l in range(801):
        Matrice3[i][l] = float(data3[i][l])
'''

def read_matrice(path,niveau):
    f = open(path)
    data = f.readlines()
    if niveau == 0:
        taille_x = 200
        taille_y = 1003
    elif niveau == 1:
        taille_x = 901
        taille_y = 1003
    elif niveau == 2:
        taille_x = 801
        taille_y = 1003
    elif niveau=='e':
        taille_x = 3
        taille_y = 1002

    matrice = np.zeros((taille_y,taille_x))
    for i in range(taille_y):
        data[i] = data[i].split()
        for j in range(taille_x):
            matrice[i][j] = float(data[i][j])
    return matrice


Matrice10_p_1 = read_matrice(fp1,0)
Matrice10_p_2 = read_matrice(fp2,1)
Matrice10_p_3 = read_matrice(fp3,2)
Matrice16_p_1 = read_matrice(fp4,0)
Matrice16_p_2 = read_matrice(fp5,1)
Matrice_e = read_matrice(fe,'e')

def energie_dep_gamma(e_inci,v,matrice10_1=Matrice10_p_1,matrice10_2=Matrice10_p_2,matrice10_3=Matrice10_p_3,matrice16_1=Matrice16_p_1,matrice16_2=Matrice16_p_2,ed=Matrice_e):
    """ 
    ----------
    Parameters
    ----------
    e_inci : float
        l'énergie incidente de particule.
    * : TYPE
        DESCRIPTION.
    matrice10_1 : matrix
        matrice de photon de 1-200keV de solution 10ml.
    matrice2 : TYPE, optional
        matrice de photon de 200-2000keV de solution 10ml.
    matrice3 : TYPE, optional
        matrice de photon de 2000-10000keV de solution 10ml.
    ed : TYPE, optional
        matrice de bins d'énergie. colonne 0: 1-200keV; colonne 1: 200-2000keV

    Returns
    -------
    result : float
        l'énergie déposée.

    """
    ## sort keV / entrée : keV
    if e_inci <= 200:
        index = int(e_inci)            # index de colonne de la matrice de l'énergie incidente la plus proche
        if v == 10: 
            matrice = matrice10_1
        elif v == 16:
            matrice = matrice16_1
        e = ed[:,0]
    
    elif e_inci <= 2000:
        index = int((e_inci-200)/2)
        if v == 10: 
            matrice = matrice10_2
        elif v == 16:
            matrice = matrice16_2
        #taille_x = 901
        e = ed[:,1]

    else:
        index = (int(e_inci)-2000)//10
        #doc = 'MCNP-MATRIX/matrice/matrice_p_2000_10000k.txt'
        matrice = matrice3
        #taille_x = 801
        e = ed[:,2]
    
    '''
    with open(doc) as f:
        data = f.readlines()
    
        matrice = np.zeros((1002,taille_x))

        for i in range(1002):
            data[i] = data[i].split()
            for j in range(taille_x):
                matrice[i][j] = float(data[i][j])
    '''
    inde = sampling(matrice[1:,index])
    if inde == 1 : result = 0
        #elif e_inci<25: result = e[inde-1]*1e3*e_inci/matrice[0][index]
    else: result = e[inde]*1e3*e_inci/matrice[0][index]
    if result  > e_inci: result = e_inci
    return result

#for i in range(10):
 #   print(energie_dep_gamma(511,16))


if absolutePath: 
    fe1 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_beta-_1_200k.txt' # electron-10ml-1-200keV-niveau 0
    fe2 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_beta-_200_2000k.txt' # electron-10ml-200-2000keV-niveau 1
    fe3 = 'G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/matrice_beta-_2000_10000k.txt' # electron-10ml-2000-10000keV-niveau 2
    fe = "G:\Python_modules\Jialin\Code\\MCNP-MATRIX/matrice/fichier/E_depose.txt"   # electron-10ml-énergie-niveau 'e'
else:
    fe1 = 'MCNP-MATRIX/matrice/fichier/matrice_beta-_1_200k.txt' # electron-10ml-1-200keV-niveau 0
    fe2 = 'MCNP-MATRIX/matrice/fichier/matrice_beta-_200_2000k.txt' # electron-10ml-200-2000keV-niveau 1
    fe3 = 'MCNP-MATRIX/matrice/fichier/matrice_beta-_2000_10000k.txt' # electron-10ml-2000-10000keV-niveau 2
    fe = "MCNP-MATRIX/matrice/fichier/E_depose.txt" # electron-10ml-énergie-niveau 'e'



Matrice10_e_1 = read_matrice(fe1,0)
Matrice10_e_2 = read_matrice(fe2,1)
Matrice10_e_3 = read_matrice(fe3,2)
#Matrice_e = read_matrice(fe,'e')


def energie_dep_beta(e_inci,*,matrice10_1=Matrice10_e_1,matrice10_2=Matrice10_e_2,matrice10_3=Matrice10_e_3,ed=Matrice_e):
    """ 
    ----------
    Parameters
    ----------
    e_inci : float
        l'énergie incidente de particule.
    * : TYPE
        DESCRIPTION.
    matrice10_1 : matrix
        matrice d'électrons de 1-200keV de solution 10ml.
    matrice2 : TYPE, optional
        matrice d'électrons de 200-2000keV de solution 10ml.
    matrice3 : TYPE, optional
        matrice d'électrons de 2000-10000keV de solution 10ml.
    ed : TYPE, optional
        matrice de bins d'énergie. colonne 0: 1-200keV; colonne 1: 200-2000keV

    Returns
    -------
    result : float
        l'énergie déposée.

    """
    ## sort keV / entrée : keV
    if e_inci <= 200:
        index = int(e_inci)            # index de colonne de la matrice de l'énergie incidente la plus proche 
        #doc = 'MCNP-MATRIX/matrice/matrice_p_1_200k.txt'
        matrice = matrice10_1
        #taille_x = 200
        e = ed[:,0]
    
    elif e_inci <= 2000:
        index = int((e_inci-200)/2)
        #doc = 'MCNP-MATRIX/matrice/matrice_p_200_2000k.txt'
        matrice = matrice10_2
        #taille_x = 901
        e = ed[:,1]

    else:
        index = (int(e_inci)-2000)//10
        #doc = 'MCNP-MATRIX/matrice/matrice_p_2000_10000k.txt'
        matrice = matrice10_3
        #taille_x = 801
        e = ed[:,2]
    
    inde = sampling(matrice[1:,index])
    if inde == 1 : result = 0
        #elif e_inci<25: result = e[inde-1]*1e3*e_inci/matrice[0][index]
    else: result = e[inde]*1e3*e_inci/matrice[0][index]
    if result  > e_inci: result = e_inci
    return result

'''
r = []   
for i in range(10):
    r.append(energie_dep_beta(17.055))
print(r)
'''
#print(Matrice_e[0:5,:])

# tic()
# energie_dep_gamma(800)
# toc()   # 0 s



## Non parametric regression

# def regress(x,y):
#     z = sm.nonparametric.lowess(y, x, return_sorted = True)
#     return z

    

def writeEffcurves(x,y,uy,rad,p,kB,SDT):
    if SDT == "S":
        file = open("EfficiencyCurves/"+''.join(rad)+"/EffS_"+''.join(rad)+'_'+''.join(str(p))+'_'+str(kB)+".txt","w")
    elif SDT == "D":
        file = open("EfficiencyCurves/"+''.join(rad)+"/EffD_"+''.join(rad)+'_'+''.join(str(p))+'_'+str(kB)+".txt","w")
    elif SDT == "T":
        file = open("EfficiencyCurves/"+''.join(rad)+"/EffT_"+''.join(rad)+'_'+''.join(str(p))+'_'+str(kB)+".txt","w")
    else:
        print("Warning: unknown profil type")
    for i, xi in enumerate(x):
        file.write(str(xi)+" "+str(y[i])+" "+str(uy[i])+"\n")
    file.close()

#======================== read ENSDF ============================================
def transf_name(rad):     #  transformer le nom de rad par exemple '11C' à 'C11'
    """
    ---------
    PARAMETRE
    ---------
    rad -- type: str par exemple '108AG'
    
    ------
    RETURN
    ------
    RAD -- type: str par exemple 'AG108' qui correspond à la structure de fille de PenNuc

    """
    # rad -- type : str par exemple '108AG'

    name_lis = re.split('(\d+)',rad)
    RAD = name_lis[2]+name_lis[1]
    return RAD

#print(transf_name('108PD'))

file_ensdf = 'decayData//All-nuclides_Ensdf.zip'
z_ensdf = zf.ZipFile(file_ensdf)
#print(z.namelist())
def readEShape(rad, *, z=z_ensdf):
    """
    --------------------------------------------------
    pour lire les fichiers dans All-nuclides_Ensdf.zip
    --------------------------------------------------
    ---------
    PARAMETRE
    ---------
    rad -- type: str par exemple 'Ag-108'
    z -- ENSDF files
    
    ------
    RETURN
    ------
    daug_name -- type: list -- les filles de désintégration
    Energy ----- type: list -- chaque élément comprend toutes les énergies de transition de la fille de même indice
    Prob ------- type: list -- chaque élément comprend toutes les proba de transition de la fille de même indice
    Type ------- type: list -- chaque élément comprend touts les types de transition de la fille de même indice

    """
       
    name = rad + '.txt'
    with z.open(name) as f:
        data = f.readlines()
        nl = np.size(data)
        #print(data)
        for i in range(nl):
            data[i] = str(data[i])
            data[i] = data[i].replace("b",'')
            data[i] = data[i].replace("\\r\\n",'')
            data[i] = data[i].replace("'",'')
            if "\\n" in data[i]:data[i] = data[i].replace("\\n","")     # pour traiter "\\n" dans Mn-52 et Mn-52m (cas particuliers)
        #print(data)
        for i in range(nl):
            data[i] = data[i].split()
        
        for i in range(nl):
            if i>0 and ('L' in data[i]) and ("AUGER" in data[i]) and ("|]" in data[i-1]):
                data.insert(i,[data[i][0],'T'])
    #print(data)
    index_auger = []
    index_end = []
    daug_name = []
    posi = []
    for i,p in enumerate(data):
        if 'DECAY' in p:
            daug_name.append(transf_name(p[0]))
        if 'Auger' in p:
            index_auger.append(i)
        if len(p)==2 and 'T' in p:
            posi.append(i)
        if 'P' in p:
            index_end.append(i)
            posi.append(i)

    Energy = []           # enregistrer les résultats (énergie) complètes
    energy = []           # enregistrer les résultats (énergie) d'une fille
    Type = []             # enregistrer les résultats (type de transition) complètes
    type_ = []            # enregistrer les résultats (type de transition) d'une fille
    Prob = []             # enregistrer les résultats (proba de transition) complètes
    prob = []             # enregistrer les résultats (proba de transition) d'une fille

    for i in range(len(posi)-1):
        start = posi[i]+1
        end = posi[i+1]
        d = data[start:end]   # bloc
        e = []                # enregistrer les résultats (énergie) d'un bloc
        prob_b = []           # enregistrer les résultats (proba) d'un bloc
        type_b = []           # enregistrer les résultats (type) d'un bloc
        if start==end:        # sauter les lignes blaches et continues
            continue
        if start-1 in index_end:   # sauter le bloc entre deux filles
            continue

        for n,p1 in enumerate(d):   
            if '-' in p1[2]:                # calculer et remplacer les intervalles
                x = p1[2].split('-') 
                p1[2] = round((float(x[0])+float(x[1]))/2,3)

            if '(total)' in d[0]:           # traiter le bloc qui comprend (total) dans la première ligne 
                if '(total)' in p1:
                    prob_b.append(float(p1[3]))
                    e.append(p1[2])
                    type_b.append(p1[-2])
                continue 
            elif '|]' in p1:                # traiter un bloc qui comprend |]
                if len(p1)>6:               # repérer la ligne qui comprend la proba
                    prob_b.append(float(p1[4]))
                e.append(float(p1[2]))      # enregistrer les valeurs d'énergie
                if 'AUGER' in p1:           # traiter le cas d'auger et |] 
                    if 'K' in p1[-2]:       # Auger K
                        type_b.append('Auger K')
                    else:
                        print('erreur')
                elif 'X' in p1[-1]:         # Rayon X
                    type_b.append(p1[-1][0:3])
            else:                           # traiter le cas sans |] ni (total)
                if len(p1)==4 and 'X' in p1[-1]:   # le cas de rayon X sans |] ni (total) ni proba
                    continue                       # sauter cette ligne
                elif len(p1)==5 and 'L' in p1:     # le cas de Auger L sans |] ni (total) ni proba
                    continue                       # sauter cette ligne
                else:                       # traiter le cas sans |] ni (total) mais complet
                    e.append(float(p1[2]))         # enregistrer énergie
                    prob_b.append(float(p1[3]))    # enregistrer proba
                    if 'L' in p1:
                        type_b.append('Auger L')   # enregistrer type Auger L
                    else:
                        type_b.append(p1[-1][0:3]) # enregistrer type Rayon X

        if len(prob_b)==1 and len(e)>1:            # calculer la valeur moyenne et l'enregistrer au cas où |] compris et valeurs complètes
            energy.append(np.mean(e))
            prob.append(prob_b[0])
            type_.append(type_b[0])
        elif len(e)==len(prob_b) and len(e)>=1:    # enregistrer les valeurs au cas où sans |] et valeurs complètes
            for i in range(len(e)):
                energy.append(e[i])
                prob.append(prob_b[i])
                type_.append(type_b[i])
        if end in index_end or end+1 in index_end: # enregistrer les résultats à la fin d'une fille
            Energy.append(energy)
            Prob.append(prob)
            Type.append(type_)
            energy = []
            prob = []
            type_ = []
    #Energy =  np.array(Energy)
    #Prob =  np.array(Prob)         
    return  daug_name,Energy,Prob,Type        

# tic()
# d,e,p,t = readEShape('Cf-252')
# toc() # 0 s
#========  tester readEShape ==============
# tic()
# d,e,p,t = readEShape('Cf-252')
#print(d,e,p,t)
#d1,e1,p1,t1 = readEShape('At-218')
#print(d1,e1,p1,t1)
# toc()
# print(d,e[0][1],p[1][2],t)
# print('  ')
# for i in range(len(d)):
#     print(d[i],e[i],p[i],t[i])
#     print(' ')
#     print(len(e[i]),len(p[i]),len(t[i]))

#============  traiter la relaxation ===============
def relaxation_atom(daugther,rad,lacune='defaut'):
    """
    ---------
    PARAMETRE
    ---------
    daugther -- type: str -- la fille tirée dans cette itération (par exemple NB95,PD110 etc.)
    rad ------- type: str -- le radionucléide étudié (par exemple Am-241, C-11 etc.) 
    lacuen ---- type: str -- la lacune atomique (par exemple 'Atom_K','Atom_L' etc.)

    ------
    RETURN
    ------
    Type ---- type de transition: Auger L ou K ou Rayon X
    Energy -- énergie correspondante

    """
    daug_name,Energy,Prob,Type = readEShape(rad)  # tirer les vecteurs de rad d'Ensdf 

    index_daug = daug_name.index(daugther)        # repérer l'indice de fille correspondante
    
    Energie = np.array(Energy[index_daug])                  # tirer le vecteur d'énergie
    probability = np.array(Prob[index_daug])                # tirer le vecteur de proba
    type_transi = Type[index_daug]                # tirer le vecteur de type


    if len(probability) > 0:                      # le cas où le vecteur de proba/energie/type n'est pas vide
        '''
        posi_L = []
        posi_K = []
        
        for i, p in enumerate(type_transi):       # repérer les indices de transition L ou K
            if 'L' in p:
                posi_L.append(i)                  # enregistrer la posotion de transition L

            elif 'K' in p:
                posi_K.append(i)                  # enregistrer la posotion de transition L
        '''
        if 'L' in lacune:                         # traiter le transition de couche L
            prob_2 = []
            energy_2 = []
            type_2 = []
            for il, pl in enumerate(type_transi):
                if 'L' in pl:
                    prob_2.append(probability[il])    # enregistrer les proba de transition L
                    energy_2.append(Energie[il])      # enregistrer les energies de transition L
                    type_2.append(type_transi[il])    # enregistrer les types de transition L

        elif 'K' in lacune:                           # traiter le transition de couche K
            prob_2 = []
            energy_2 = []
            type_2 = []
            for ik, pk in enumerate(type_transi):     
                if 'K' in pk:
                    prob_2.append(probability[ik])    # enregistrer les proba de transition K
                    energy_2.append(Energie[ik])      # enregistrer les energie de transition K
                    type_2.append(type_transi[ik])    # enregistrer les type de transition K

        #elif lacune=='defaut':                        # traiter le cas particulier qui ne précise pas la lacune
            #prob_2 = probability
            #energy_2 = Energie
            #type_2 = type_transi
            
        else: # try to debugg
            # print("issue: ", lacune)
            prob_2 = 0   #probability
            energy_2 = 0  # Energie
            #if "M" in lacune:
                #type_2 = "Atom_M"  #type_transi
            #if "N" in lacune:
                #type_2 = "Atom_N"            
        
     # sampling
        #if len(probability)>1:               # le cas où la taille du vecteur de proba supérieur à 1

            #prob_somme = np.sum(prob_2)      # calculer la somme de proba
            #prob_2 /= prob_somme             # normaliser la proba
        if lacune != "Atom_M" and lacune != "Atom_N":
            if len(prob_2) != 0:
                prob_2 = np.array(prob_2)           # convert to array
                if len(probability)>1:
                    prob_somme = np.sum(prob_2)      # calculer la somme de proba
                    prob_2 /= prob_somme 
                index_fin = sampling(prob_2)        # sample in probability of transition
                type_fin = type_2[index_fin]        # type of transition     
                energie_fin = energy_2[index_fin]   # energy of the transition
            else:
                # print("pas de transition de rayon X ni d'électron Auger pour cette lacune: ",lacune)
                type_fin = 'NON'
                energie_fin = 0            
        else:
            type_fin = 'NON'
            energie_fin = 0 
    else:                                            # le cas où le vecteur de proba est vide 
        #print("pas de transition de rayon X ni d'électron Auger")
        type_fin = 'NON'
        energie_fin = 0
    return type_fin,energie_fin
# tic()
# tf,ef = relaxation_atom('CR52', 'Mn-52', 'Atom_K')
# toc() # 0 s
#print(tf,ef)
